import os
from scipy.signal import medfilt
from astrodash.preprocessing import ReadSpectrumFile, ProcessingTools, PreProcessSpectrum
from astrodash.array_tools import zero_non_overlap_part, normalise_spectrum


class CombineSnAndHost(object):
    def __init__(self, snInfo, galInfo, w0, w1, nw):
        self.snInfo = snInfo
        self.galInfo = galInfo
        self.processingTools = ProcessingTools()
        self.numSplinePoints = 13
        self.preProcess = PreProcessSpectrum(w0, w1, nw)

    def overlapped_spectra(self):
        snWave, snFlux, snMinIndex, snMaxIndex = self.snInfo
        galWave, galFlux, galMinIndex, galMaxIndex = self.galInfo

        minIndex = max(snMinIndex, galMinIndex)
        maxIndex = min(snMaxIndex, galMaxIndex)

        snFlux = zero_non_overlap_part(snFlux, minIndex, maxIndex)
        galFlux = zero_non_overlap_part(galFlux, minIndex, maxIndex)

        return snWave, snFlux, galWave, galFlux, minIndex, maxIndex

    def sn_plus_gal(self, snCoeff, galCoeff):
        snWave, snFlux, galWave, galFlux, minIndex, maxIndex = self.overlapped_spectra()

        combinedFlux = (snCoeff * snFlux) + (galCoeff * galFlux)

        return snWave, combinedFlux, minIndex, maxIndex

    def template_data(self, snCoeff, galCoeff, z):
        wave, flux, minIndex, maxIndex = self.sn_plus_gal(snCoeff, galCoeff)
        wave, flux = self.processingTools.redshift_spectrum(wave, flux, z)
        flux = zero_non_overlap_part(flux, minIndex, maxIndex, outerVal=0)
        binnedWave, binnedFlux, minIndex, maxIndex = self.preProcess.log_wavelength(wave[minIndex:maxIndex + 1],
                                                                                    flux[minIndex:maxIndex + 1])
        newFlux, continuum = self.preProcess.continuum_removal(binnedWave, binnedFlux, self.numSplinePoints, minIndex,
                                                               maxIndex)
        meanZero = self.preProcess.mean_zero(newFlux, minIndex, maxIndex)
        apodized = self.preProcess.apodize(meanZero, minIndex, maxIndex)
        fluxNorm = normalise_spectrum(apodized)
        fluxNorm = zero_non_overlap_part(fluxNorm, minIndex, maxIndex, outerVal=0.5)
        # Could  median filter here, but trying without it now

        return binnedWave, fluxNorm, (minIndex, maxIndex)


class BinTemplate(object):
    def __init__(self, filename, templateType, w0, w1, nw):
        self.w0 = w0
        self.w1 = w1
        self.nw = nw
        self.numSplinePoints = 13
        self.filename = os.path.basename(filename)
        self.templateType = templateType
        self.preProcess = PreProcessSpectrum(w0, w1, nw)
        self.readSpectrumFile = ReadSpectrumFile(filename, w0, w1, nw)
        self.spectrum = self.readSpectrumFile.file_extension(template=True)
        if templateType == 'sn':
            if len(self.spectrum) == 6:
                self.snidTemplate = True
                self.wave, self.fluxes, self.nCols, self.ages, self.tType, self.splineInfo = self.spectrum
            else:
                self.snidTemplate = False
                self.wave, self.flux, self.nCols, self.ages, self.tType = self.spectrum

        elif templateType == 'gal':
            self.wave, self.flux = self.spectrum
            self.tType = self.filename
        else:
            print("INVALID ARGUMENT FOR TEMPLATE TYPE")

    def bin_template(self, ageIdx=None):
        if self.templateType == 'sn':
            if ageIdx is None and self.snidTemplate:
                print("AGE INDEX ARGUMENT MISSING")
                return None
            else:
                return self._bin_sn_template(ageIdx)

        elif self.templateType == 'gal':
            return self._bin_gal_template()
        else:
            print("INVALID ARGUMENT FOR TEMPLATE TYPE")
            return None

    def _bin_sn_template(self, ageIdx):
        # Undo continuum in the following step in preprocessing.py
        if self.snidTemplate:
            wave, flux = self.wave, self.fluxes[
                ageIdx]  # self.snReadSpectrumFile.snid_template_undo_processing(self.snWave, self.snFluxes[ageIdx], self.splineInfo, ageIdx)
            binnedWave, binnedFlux, minIndex, maxIndex = self.preProcess.log_wavelength(wave, flux)
            binnedFluxNorm = normalise_spectrum(binnedFlux)
            binnedFluxNorm = zero_non_overlap_part(binnedFluxNorm, minIndex, maxIndex, outerVal=0.5)
        else:
            wave, flux = self.wave, medfilt(self.flux, kernel_size=3)
            flux = normalise_spectrum(flux)
            binnedWave, binnedFlux, minIndex, maxIndex = self.preProcess.log_wavelength(wave, flux)
            contRemovedFlux, continuum = self.preProcess.continuum_removal(binnedWave, binnedFlux, self.numSplinePoints,
                                                                           minIndex, maxIndex)
            meanzero = self.preProcess.mean_zero(contRemovedFlux, minIndex, maxIndex)
            apodized = self.preProcess.apodize(meanzero, minIndex, maxIndex)
            binnedFluxNorm = normalise_spectrum(apodized)
            binnedFluxNorm = zero_non_overlap_part(binnedFluxNorm, minIndex, maxIndex, outerVal=0.5)

        return binnedWave, binnedFluxNorm, minIndex, maxIndex

    def _bin_gal_template(self):
        wave, flux = self.readSpectrumFile.two_col_input_spectrum(self.wave, self.flux, z=0)
        # flux = normalise_spectrum(flux)
        binnedWave, binnedFlux, minIndex, maxIndex = self.preProcess.log_wavelength(wave, flux)
        contRemovedFlux, continuum = self.preProcess.continuum_removal(binnedWave, binnedFlux, self.numSplinePoints,
                                                                       minIndex, maxIndex)
        newFlux = contRemovedFlux * continuum  # Spectral features weighted by the continuum
        fluxNorm = normalise_spectrum(newFlux)
        fluxNorm = zero_non_overlap_part(fluxNorm, minIndex, maxIndex, outerVal=0.5)

        return binnedWave, fluxNorm, minIndex, maxIndex


def training_template_data(snAgeIdx, snCoeff, galCoeff, z, snFile, galFile, w0, w1, nw):
    snBinTemplate = BinTemplate(filename=snFile, templateType='sn', w0=w0, w1=w1, nw=nw)
    snInfo = snBinTemplate.bin_template(snAgeIdx)

    if galFile is None:  # No host combining
        binnedWave, fluxNorm, minIndex, maxIndex = snInfo
    else:
        galBinTemplate = BinTemplate(filename=galFile, templateType='gal', w0=w0, w1=w1, nw=nw)
        galInfo = galBinTemplate.bin_template()
        combineSnAndHost = CombineSnAndHost(snInfo, galInfo, w0, w1, nw)
        binnedWave, fluxNorm, (minIndex, maxIndex) = combineSnAndHost.template_data(snCoeff, galCoeff, z)

    nCols, ages, tType = snBinTemplate.nCols, snBinTemplate.ages, snBinTemplate.tType

    return binnedWave, fluxNorm, minIndex, maxIndex, nCols, ages, tType



import os
import sys
import numpy as np
from PyQt5 import QtGui
from PyQt5.QtCore import QThread, pyqtSignal
import pyqtgraph as pg
from astrodash.design import Ui_MainWindow
from astrodash.restore_model import LoadInputSpectra, BestTypesListSingleRedshift, get_training_parameters, \
    classification_split
from astrodash.create_arrays import AgeBinning
from astrodash.read_binned_templates import load_templates, get_templates, combined_sn_and_host_data
from astrodash.false_positive_rejection import combined_prob, RlapCalc
from astrodash.calculate_redshift import get_median_redshift, get_redshift_axis
from astrodash.read_from_catalog import catalogDict


class MainApp(QtGui.QMainWindow, Ui_MainWindow):
    def __init__(self, parent=None, inputFilename="DefaultFilename", data_files='models_v06'):
        super(MainApp, self).__init__(parent)
        self.setupUi(self)

        self.mainDirectory = os.path.dirname(os.path.abspath(__file__))
        self.data_files = data_files
        self.templates()
        self.plotted = False
        self.plotZ = 0
        self.hostFraction = 0
        self.inputFluxes = np.zeros((2, int(self.nw)))
        self.inputImageUnRedshifted = np.zeros(int(self.nw))
        self.templatePlotFlux = np.zeros(int(self.nw))
        self.templateSubIndex = 0
        self.snName = 'Ia-norm'
        self.snAge = '-20 to -18'
        self.hostName = 'No Host'
        self.lineEditKnownZ.setText('')
        self.infLine = pg.InfiniteLine(self.plotZ, pen={'width': 3, 'color': (135, 206, 250)}, movable=True,
                                       bounds=[-1, 1], hoverPen={'color': (255, 0, 0), 'width': 3}, label='z',
                                       labelOpts={'color': (135, 206, 250), 'position': 0.2})
        self.infLine.sigPositionChanged.connect(self.cross_corr_redshift_changed)

        self.pushButtonLeftTemplate.clicked.connect(self.select_sub_template_left)
        self.pushButtonRightTemplate.clicked.connect(self.select_sub_template_right)
        self.btnBrowse.clicked.connect(self.select_input_file)
        self.listWidget.itemClicked.connect(self.list_item_clicked)
        self.btnRefit.clicked.connect(self.fit_spectra)
        self.inputFilename = inputFilename
        self.inputFilenameSetText = inputFilename
        self.progressBar.setValue(100)
        self.add_combo_box_entries()
        self.labelRlapScore.setText('')

        self.select_tensorflow_model()
        self.checkBoxKnownZ.stateChanged.connect(self.select_tensorflow_model)
        self.checkBoxClassifyHost.stateChanged.connect(self.select_tensorflow_model)

        self.lineEditInputFilename.textChanged.connect(self.input_filename_changed)

        self.horizontalSliderSmooth.valueChanged.connect(self.smooth_slider_changed)
        self.lineEditSmooth.textChanged.connect(self.smooth_text_changed)

        self.horizontalSliderRedshift.valueChanged.connect(self.redshift_slider_changed)
        self.lineEditRedshift.textChanged.connect(self.redshift_text_changed)

        self.horizontalSliderHostFraction.valueChanged.connect(self.host_fraction_slider_changed)
        self.lineEditHostFraction.editingFinished.connect(self.host_fraction_text_changed)

        self.comboBoxSNType.currentIndexChanged.connect(self.combo_box_changed)
        self.comboBoxAge.currentIndexChanged.connect(self.combo_box_changed)
        self.comboBoxHost.currentIndexChanged.connect(self.combo_box_changed)

        self.btnQuit.clicked.connect(self.close)

        self.graphicsView.plotItem.layout.removeItem(self.graphicsView.plotItem.getAxis('top'))
        self.cAxis = pg.AxisItem(orientation='top', parent=self.graphicsView.plotItem)
        self.cAxis.linkToView(self.graphicsView.plotItem.vb)
        self.graphicsView.plotItem.axes['top']['item'] = self.cAxis
        self.graphicsView.plotItem.layout.addItem(self.cAxis, 1, 1)

    def select_tensorflow_model(self):
        if self.checkBoxKnownZ.isChecked():
            self.lineEditKnownZ.setEnabled(True)
            if self.checkBoxClassifyHost.isChecked():
                self.modelFilename = os.path.join(self.mainDirectory, self.data_files,
                                                  "models/zeroZ_classifyHost/tensorflow_model.ckpt")
            else:
                self.modelFilename = os.path.join(self.mainDirectory, self.data_files,
                                                  "models/zeroZ/tensorflow_model.ckpt")
        else:
            self.lineEditKnownZ.setEnabled(False)
            if self.checkBoxClassifyHost.isChecked():
                self.modelFilename = os.path.join(self.mainDirectory, self.data_files,
                                                  "models/agnosticZ_classifyHost/tensorflow_model.ckpt")
            else:
                self.modelFilename = os.path.join(self.mainDirectory, self.data_files,
                                                  "models/agnosticZ/tensorflow_model.ckpt")
        if not os.path.isfile(self.modelFilename + ".index"):
            QtGui.QMessageBox.critical(self, "Error", "Model does not exist")

    def templates(self):
        pars = get_training_parameters()
        self.w0, self.w1, self.minAge, self.maxAge, self.ageBinSize, self.typeList, self.nTypes, self.nw, self.hostTypes \
            = pars['w0'], pars['w1'], pars['minAge'], pars['maxAge'], pars['ageBinSize'], pars['typeList'], pars[
            'nTypes'], pars['nw'], pars['galTypeList']

        self.dwlog = np.log(self.w1 / self.w0) / self.nw
        self.wave = self.w0 * np.exp(np.arange(0, self.nw) * self.dwlog)

        self.snTemplates, self.galTemplates = load_templates(
            os.path.join(self.mainDirectory, self.data_files, 'models/sn_and_host_templates.npz'))

    def get_sn_and_host_templates(self, snName, snAge, hostName):
        snInfos, snNames, hostInfos, hostNames = get_templates(snName, snAge, hostName, self.snTemplates,
                                                               self.galTemplates, self.nw)

        return snInfos, snNames, hostInfos, hostNames

    def get_template_info(self):  #
        snInfos, snNames, hostInfos, hostNames = self.get_sn_and_host_templates(self.snName, self.snAge, self.hostName)
        numOfSubTemplates = len(snNames)
        if self.templateSubIndex >= numOfSubTemplates:
            self.templateSubIndex = 0
        if self.templateSubIndex < 0:
            self.templateSubIndex = numOfSubTemplates - 1

        if snInfos != []:
            name = "%s_%s" % (snNames[self.templateSubIndex], hostNames[0])
            if self.hostName != "No Host":
                snCoeff = 1 - self.hostFraction / 100.
                galCoeff = self.hostFraction / 100.
            else:
                snCoeff = 1
                galCoeff = 0
            wave, flux, minMaxIndex = combined_sn_and_host_data(snCoeff=snCoeff, galCoeff=galCoeff, z=0,
                                                                snInfo=snInfos[self.templateSubIndex],
                                                                galInfo=hostInfos[0], w0=self.w0, w1=self.w1,
                                                                nw=self.nw)
            return flux, name, minMaxIndex
        else:
            flux = np.zeros(self.nw)
            name = "NO_TEMPLATES!"
            minMaxIndex = (0, 0)

        return flux, name, minMaxIndex

    def select_sub_template_right(self):
        self.templateSubIndex += 1
        self.plot_sub_templates()

    def select_sub_template_left(self):
        self.templateSubIndex -= 1
        self.plot_sub_templates()

    def plot_sub_templates(self):
        self.templatePlotFlux, self.templatePlotName, self.templateMinMaxIndex = self.get_template_info()
        print(self.templatePlotName)
        self.plot_best_matches()
        self.plot_cross_corr()

    def combo_box_changed(self):
        self.snName = str(self.comboBoxSNType.currentText())
        self.snAge = str(self.comboBoxAge.currentText())
        self.hostName = str(self.comboBoxHost.currentText())

        self.redshift, self.crossCorrs, self.medianName, self.redshiftErr = self.calc_redshift(self.snName, self.snAge)
        self.set_template_sub_index(self.medianName)
        self.templatePlotFlux, self.templatePlotName, self.templateMinMaxIndex = self.get_template_info()
        self.plot_cross_corr()
        if self.knownRedshift:
            self.plot_best_matches()
        else:
            self.set_plot_redshift(self.redshift)

    def add_combo_box_entries(self):
        ageLabels = AgeBinning(self.minAge, self.maxAge, self.ageBinSize).age_labels()
        for i in range(len(ageLabels)):
            self.comboBoxAge.addItem(ageLabels[i])

        for typeName in self.typeList:
            self.comboBoxSNType.addItem(typeName)

        self.comboBoxHost.addItem("No Host")
        for hostName in self.hostTypes:
            self.comboBoxHost.addItem(hostName)

    def host_fraction_slider_changed(self):
        self.hostFraction = self.horizontalSliderHostFraction.value()
        self.lineEditHostFraction.setText("%s%%" % str(self.hostFraction))
        self.templatePlotFlux, self.templatePlotName, self.templateMinMaxIndex = self.get_template_info()
        self.plot_best_matches()

    def host_fraction_text_changed(self):
        try:
            self.hostFraction = float(self.lineEditHostFraction.text().strip("%%"))
            self.horizontalSliderHostFraction.setValue(int(self.hostFraction))
        except ValueError:
            print("Host Fraction Value Error")

    def redshift_slider_changed(self):
        self.plotZ = self.horizontalSliderRedshift.value() / 10000.
        self.lineEditRedshift.setText(str(round(self.plotZ, 3)))
        self.infLine.setValue(self.plotZ)

    def redshift_text_changed(self):
        try:
            self.plotZ = float(self.lineEditRedshift.text())
            self.horizontalSliderRedshift.setValue(int(self.plotZ * 10000))
            self.infLine.setValue(self.plotZ)
            self.plot_best_matches()
        except ValueError:
            print("Redshift Value Error")

    def cross_corr_redshift_changed(self):
        self.plotZ = self.infLine.value()
        self.horizontalSliderRedshift.setValue(int(self.plotZ * 10000))
        self.lineEditRedshift.setText(str(round(self.plotZ, 3)))

    def set_plot_redshift(self, plotZ):
        plotZ = float(plotZ)
        self.plotZ = plotZ
        self.lineEditRedshift.setText(str(round(self.plotZ, 3)))
        self.horizontalSliderRedshift.setValue(int(plotZ * 10000))
        self.infLine.setValue(self.plotZ)
        self.plot_best_matches()

    def smooth_slider_changed(self):
        self.lineEditSmooth.setText(str(self.horizontalSliderSmooth.value()))

    def smooth_text_changed(self):
        try:
            self.horizontalSliderSmooth.setValue(int(self.lineEditSmooth.text()))
        except ValueError:
            pass

    def input_filename_changed(self):
        inputFilename = self.lineEditInputFilename.text()
        if inputFilename == self.inputFilenameSetText or inputFilename == self.inputFilename or inputFilename == "":
            pass
        else:
            self.inputFilename = inputFilename

    def select_input_file(self):
        inputFilename = QtGui.QFileDialog.getOpenFileName(self, "Select a spectrum file")[0]
        if (inputFilename == self.inputFilename) or (inputFilename == ""):
            pass
        else:
            self.inputFilename = inputFilename
            self.inputFilenameSetText = self.inputFilename  # os.path.basename(self.inputFilename)
            self.lineEditInputFilename.setText(self.inputFilenameSetText)

            self.fit_spectra()

    def fit_spectra(self):
        self.cancelledFitting = False
        try:
            self.smooth = int(self.lineEditSmooth.text())
        except ValueError:
            QtGui.QMessageBox.critical(self, "Error", "Smooth must be positive integer")
            return
        try:
            self.minWave = int(self.lineEditMinWave.text())
            self.maxWave = int(self.lineEditMaxWave.text())
        except ValueError:
            QtGui.QMessageBox.critical(self, "Error",
                                       "Min and max waves must be integers between %d and %d" % (self.w0, self.w1))
            return
        if self.checkBoxClassifyHost.isChecked():
            self.classifyHost = True
        else:
            self.classifyHost = False
        if self.checkBoxKnownZ.isChecked():
            self.knownRedshift = True
            try:
                knownZ = float(self.lineEditKnownZ.text())
                self.bestRedshift = knownZ
                self.bestRedshiftErr = None
            except ValueError:
                QtGui.QMessageBox.critical(self, "Error", "Enter Known Redshift")
                return
        else:
            self.knownRedshift = False
            knownZ = 0
            self.lineEditKnownZ.setText("")
        if not os.path.isfile(self.modelFilename + ".index"):
            QtGui.QMessageBox.critical(self, "Error", "Model does not exist")
            return
        if not isinstance(self.inputFilename, (list, np.ndarray)) and not hasattr(self.inputFilename,
                                                                                  'read') and not os.path.isfile(
                self.inputFilename) and self.inputFilename.split('-')[0] not in list(
                catalogDict.keys()):  # Not an array and not a file-handle and not a file and not a catalog input
            QtGui.QMessageBox.critical(self, "Error", "File not found!")
            return
        try:
            if self.inputFilename.split('-')[0] in list(catalogDict.keys()):
                knownZ = 0
                self.bestRedshift = 0
                self.bestRedshiftErr = None
                self.lineEditKnownZ.setText("")
        except Exception as e:
            pass  # Not an Open Supernova catalog object. Make this neater in future.

        if self.checkBoxRlap.isChecked():
            self.getRlapScores = True
        else:
            self.getRlapScores = False

        self.progressBar.setValue(36)
        self.set_plot_redshift(knownZ)

        self.fitThread = FitSpectrumThread(self.inputFilename, knownZ, self.modelFilename, self.smooth,
                                           self.classifyHost, self.minWave, self.maxWave)
        self.fitThread.trigger.connect(self.load_spectrum_single_redshift)

        self.fitThread.start()

        self.btnCancel.clicked.connect(self.cancel)

    def cancel(self):
        if not self.cancelledFitting:
            self.cancelledFitting = True
            self.fitThread.terminate()
            self.progressBar.setValue(100)
            QtGui.QMessageBox.information(self, "Cancelled!", "Stopped Fitting Input Spectrum")

    def load_spectrum_single_redshift(self, spectrumInfo):
        self.bestTypes, self.softmax, self.idx, self.typeNamesList, self.inputImageUnRedshifted, self.inputMinMaxIndex = spectrumInfo
        self.progressBar.setValue(85)  # self.progressBar.value()+)
        self.done_fit_thread_single_redshift()

    def done_fit_thread_single_redshift(self):
        if not self.cancelledFitting:
            self.plotted = True
            self.list_best_matches_single_redshift()
            self.set_plot_redshift(self.bestRedshift)
            self.plot_cross_corr()
            self.progressBar.setValue(100)
            # QtGui.QMessageBox.information(self, "Done!", "Finished Fitting Input Spectrum")

    def best_broad_type(self):
        bestMatchList = []
        for i in range(10):
            host, name, age = classification_split(self.bestTypes[i])
            bestMatchList.append([host, name, age, self.softmax[i]])
        host, prevName, bestAge, probTotal, reliableFlag = combined_prob(bestMatchList)
        self.labelBestSnType.setText(prevName)
        self.labelBestAgeRange.setText(bestAge)
        self.labelBestHostType.setText(host)
        if host == "":
            self.labelBestHostType.setFixedWidth(0)
        if self.bestRedshiftErr is None:
            self.labelBestRedshift.setText(str(self.bestRedshift))
        else:
            self.labelBestRedshift.setText("{} {} {}".format(str(self.bestRedshift), "±", self.bestRedshiftErr))
        self.labelBestRelProb.setText("%s%%" % str(round(100 * probTotal, 2)))
        if host == "":
            self.labelBestHostType.setFixedWidth(0)
        if reliableFlag:
            self.labelInconsistentWarning.setText("Reliable matches")
            self.labelInconsistentWarning.setStyleSheet('color: green')
        else:
            self.labelInconsistentWarning.setText("Unreliable matches")
            self.labelInconsistentWarning.setStyleSheet('color: red')

    def get_smoothed_templates(self, snName, snAge, hostName):
        snInfos, snNames, hostInfos, hostNames = self.get_sn_and_host_templates(snName, snAge, hostName)
        fluxes = []
        minMaxIndexes = []
        for i in range(len(snNames)):
            wave, flux, minMaxIndex = combined_sn_and_host_data(snCoeff=1, galCoeff=0, z=0, snInfo=snInfos[i],
                                                                galInfo=hostInfos[0], w0=self.w0, w1=self.w1,
                                                                nw=self.nw)

            fluxes.append(flux)
            minMaxIndexes.append(minMaxIndex)

        return fluxes, snNames, minMaxIndexes

    def low_rlap_warning_label(self, bestName, bestAge, bestHost):
        fluxes, snNames, templateMinMaxIndexes = self.get_smoothed_templates(bestName, bestAge, bestHost)
        rlapCalc = RlapCalc(self.inputImageUnRedshifted, fluxes, snNames, self.wave, self.inputMinMaxIndex,
                            templateMinMaxIndexes)
        rlapLabel, rlapWarning = rlapCalc.rlap_label()

        return rlapLabel, rlapWarning

    def list_best_matches_single_redshift(self):
        print("listing best matches...")
        redshifts, redshiftErrs = self.best_redshifts()
        self.listWidget.clear()

        header = ['No.', 'Type', 'Age', 'Softmax Prob.']
        if self.classifyHost:
            header.insert(1, 'Host')
        if not self.knownRedshift:
            header.insert(3, 'Redshift')
        if self.getRlapScores:
            header.insert(5, 'rlap')
        self.listWidget.addItem("".join(word.ljust(25) for word in header))

        for i in range(20):
            host, name, age = classification_split(self.bestTypes[i])
            prob = self.softmax[i]
            redshift = redshifts[i]

            line = [str(i + 1), name, age, str(prob)]
            if self.classifyHost:
                line.insert(1, host)
            if not self.knownRedshift:
                line.insert(3, str(redshift))
            if self.getRlapScores:
                fluxes, snNames, templateMinMaxIndexes = self.get_smoothed_templates(name, age, host)
                rlapCalc = RlapCalc(self.inputImageUnRedshifted, fluxes, snNames, self.wave, self.inputMinMaxIndex,
                                    templateMinMaxIndexes)
                rlap = rlapCalc.rlap_label()[0]
                line.insert(5, str(rlap))
            self.listWidget.addItem("".join(word.ljust(25) for word in line))

            if i == 0:
                SNTypeComboBoxIndex = self.comboBoxSNType.findText(name)
                self.comboBoxSNType.setCurrentIndex(SNTypeComboBoxIndex)
                AgeComboBoxIndex = self.comboBoxAge.findText(age)
                self.comboBoxAge.setCurrentIndex(AgeComboBoxIndex)
                hostComboBoxIndex = self.comboBoxHost.findText(host)
                self.comboBoxHost.setCurrentIndex(hostComboBoxIndex)

                rlap, rlapWarning = self.low_rlap_warning_label(name, age, host)
                if rlapWarning:
                    self.labelRlapWarning.setText("Low rlap: {0}".format(rlap))
                    self.labelRlapWarning.setStyleSheet('color: red')
                else:
                    self.labelRlapWarning.setText("Good rlap: {0}".format(rlap))
                    self.labelRlapWarning.setStyleSheet('color: green')

            if not self.knownRedshift:
                self.bestRedshift = redshifts[0]
                self.bestRedshiftErr = redshiftErrs[0]
        self.best_broad_type()

    def list_item_clicked(self, item):
        if item.text()[0].isdigit():
            host = "No Host"
            if self.knownRedshift:
                if self.classifyHost:
                    if self.getRlapScores:
                        index, host, snTypePlot, age1, to, age3, softmax, rlap = str(item.text()).split()
                    else:
                        index, host, snTypePlot, age1, to, age3, softmax = str(item.text()).split()
                else:
                    if self.getRlapScores:
                        index, snTypePlot, age1, to, age3, softmax, rlap = str(item.text()).split()
                    else:
                        index, snTypePlot, age1, to, age3, softmax = str(item.text()).split()
            else:
                if self.classifyHost:
                    if self.getRlapScores:
                        index, host, snTypePlot, age1, to, age3, redshift, softmax, rlap = str(item.text()).split()
                    else:
                        index, host, snTypePlot, age1, to, age3, redshift, softmax = str(item.text()).split()
                else:
                    if self.getRlapScores:
                        index, snTypePlot, age1, to, age3, redshift, softmax, rlap = str(item.text()).split()
                    else:
                        index, snTypePlot, age1, to, age3, redshift, softmax = str(item.text()).split()
                self.set_plot_redshift(redshift)
            agePlot = age1 + ' to ' + age3

            self.plot_cross_corr()
            snTypeComboBoxIndex = self.comboBoxSNType.findText(snTypePlot)
            self.comboBoxSNType.setCurrentIndex(snTypeComboBoxIndex)
            AgeComboBoxIndex = self.comboBoxAge.findText(agePlot)
            self.comboBoxAge.setCurrentIndex(AgeComboBoxIndex)
            hostComboBoxIndex = self.comboBoxHost.findText(host)
            self.comboBoxHost.setCurrentIndex(hostComboBoxIndex)

    def plot_best_matches(self):
        if self.plotted:
            templateWave = self.wave * (1 + (self.plotZ))
            self.labelTemplateName.setText(self.templatePlotName)

            self.graphicsView.clear()
            inputPlotFlux = self.inputImageUnRedshifted
            self.graphicsView.plot(self.wave, inputPlotFlux, name='Input Spectrum', pen={'color': (0, 255, 0)})
            self.graphicsView.plot(templateWave, self.templatePlotFlux, name=self.templatePlotName,
                                   pen={'color': (255, 0, 0)})
            self.graphicsView.setXRange(int(self.w0), int(self.w1))
            self.graphicsView.setYRange(0, 1)
            self.graphicsView.plotItem.showGrid(x=True, y=True, alpha=0.95)
            self.graphicsView.plotItem.setLabels(bottom="Observed Wavelength (<font>&#8491;</font>)")

            try:
                self.cAxis.setScale(1 / (1 + self.plotZ))
            except ZeroDivisionError:
                print("Invalid redshift. Redshift cannot be -1.")
            self.cAxis.setGrid(False)
            self.cAxis.setLabel("Rest Wavelength (<font>&#8491;</font>)")

            if np.any(self.templatePlotFlux):
                rlapCalc = RlapCalc(self.inputImageUnRedshifted, [self.templatePlotFlux], [self.templatePlotName],
                                    self.wave, self.inputMinMaxIndex, [self.templateMinMaxIndex])
                rlap = rlapCalc.rlap_label()[0]
                self.labelRlapScore.setText("rlap: {0}".format(rlap))

    def best_redshifts(self):
        redshifts = []
        redshiftErrs = []
        for i in range(20):
            host, name, age = classification_split(self.bestTypes[i])
            redshift, crossCorr, medianName, redshiftErr = self.calc_redshift(name, age)
            redshifts.append(redshift)
            redshiftErrs.append(redshiftErr)
        return redshifts, redshiftErrs

    def set_template_sub_index(self, templateName):
        snInfos, snNames, hostInfos, hostNames = self.get_sn_and_host_templates(self.snName, self.snAge, self.hostName)
        if snNames.size:
            self.templateSubIndex = np.where(snNames == templateName)[0][0]

    def calc_redshift(self, snName, snAge):
        host = "No Host"
        snInfos, snNames, hostInfos, hostNames = self.get_sn_and_host_templates(snName, snAge, host)
        numOfSubTemplates = len(snNames)
        templateNames = snNames
        templateFluxes = []
        templateMinMaxIndexes = []
        for i in range(numOfSubTemplates):
            templateFluxes.append(snInfos[i][1])
            templateMinMaxIndexes.append((snInfos[i][2], snInfos[i][3]))

        redshift, crossCorrs, medianName, redshiftErr = get_median_redshift(self.inputImageUnRedshifted, templateFluxes,
                                                                            self.nw, self.dwlog, self.inputMinMaxIndex,
                                                                            templateMinMaxIndexes, templateNames,
                                                                            outerVal=0.5)
        if redshift is None:
            return 0, 0, "", 0

        return round(redshift, 4), crossCorrs, medianName, round(redshiftErr, 4)

    def plot_cross_corr(self):
        zAxis = get_redshift_axis(self.nw, self.dwlog)
        if type(self.crossCorrs) == dict:
            crossCorr = np.real(self.crossCorrs["_".join(self.templatePlotName.split('_')[:-1])])
        else:
            crossCorr = np.zeros(self.nw)
        self.graphicsView_2.clear()
        self.graphicsView_2.plot(zAxis, crossCorr)
        self.graphicsView_2.setXRange(0, 1)
        self.graphicsView_2.setYRange(min(crossCorr), max(crossCorr))
        self.graphicsView_2.plotItem.showGrid(x=True, y=True, alpha=0.95)
        # self.graphicsView_2.plotItem.setLabels(bottom="z")
        self.graphicsView_2.addItem(self.infLine)

    def browse_folder(self):
        self.listWidget.clear()
        directory = QtGui.QFileDialog.getExistingDirectory(self, "Pick a folder")

        if directory:
            for file_name in os.listdir(directory):
                self.listWidget.addItem(file_name)


class FitSpectrumThread(QThread):
    trigger = pyqtSignal('PyQt_PyObject')

    def __init__(self, inputFilename, knownZ, modelFilename, smooth, classifyHost, minWave, maxWave):
        QThread.__init__(self)
        self.inputFilename = inputFilename
        self.knownZ = knownZ
        self.modelFilename = modelFilename
        self.smooth = smooth
        self.classifyHost = classifyHost
        self.minWave = minWave
        self.maxWave = maxWave

    def __del__(self):
        self.wait()

    def _input_spectrum_single_redshift(self):
        trainParams = get_training_parameters()
        loadInputSpectraUnRedshifted = LoadInputSpectra(self.inputFilename, 0, self.smooth, trainParams, self.minWave,
                                                        self.maxWave, self.classifyHost)
        inputImageUnRedshifted, inputRedshift, typeNamesList, nw, nBins, minMaxIndexUnRedshifted = loadInputSpectraUnRedshifted.input_spectra()

        loadInputSpectra = LoadInputSpectra(self.inputFilename, self.knownZ, self.smooth, trainParams, self.minWave,
                                            self.maxWave, self.classifyHost)
        inputImage, inputRedshift, typeNamesList, nw, nBins, minMaxIndex = loadInputSpectra.input_spectra()
        bestTypesList = BestTypesListSingleRedshift(self.modelFilename, inputImage, typeNamesList, nw, nBins)
        bestTypes = bestTypesList.bestTypes[0]
        softmax = bestTypesList.softmaxOrdered[0]
        idx = bestTypesList.idx[0]

        return bestTypes, softmax, idx, typeNamesList, inputImageUnRedshifted[0], minMaxIndexUnRedshifted[0]

    def run(self):
        spectrumInfo = self._input_spectrum_single_redshift()
        self.trigger.emit(spectrumInfo)


def main():
    app = QtGui.QApplication(sys.argv)
    sys._excepthook = sys.excepthook

    # def exception_hook(exctype, value, traceback):
    #     print(exctype, value, traceback)
    #     sys._excepthook(exctype, value, traceback)
    #     QtGui.QMessageBox.critical("Error", "Unintended exception")
    #
    # sys.excepthook = exception_hook

    form = MainApp()
    form.show()
    app.exec_()


if __name__ == '__main__':
    main()



import numpy as np


def labels_indexes_to_arrays(labelsIndexes, nLabels):
    numArrays = len(labelsIndexes)
    labelsIndexes = labelsIndexes
    labels = np.zeros((numArrays, nLabels))
    labels[np.arange(numArrays), labelsIndexes] = 1

    return labels


def zero_non_overlap_part(array, minIndex, maxIndex, outerVal=0.):
    slicedArray = np.copy(array)
    slicedArray[0:minIndex] = outerVal * np.ones(minIndex)
    slicedArray[maxIndex:] = outerVal * np.ones(len(array) - maxIndex)

    return slicedArray


def normalise_spectrum(flux):
    if len(flux) == 0 or min(flux) == max(flux):  # No data
        fluxNorm = np.zeros(len(flux))
    else:
        fluxNorm = (flux - min(flux)) / (max(flux) - min(flux))

    return fluxNorm


def mean_zero_spectra(flux, minIndex, maxIndex, nw):
    fluxOut = np.copy(flux)
    fluxOut[0:minIndex] = np.full(minIndex, fluxOut[minIndex])
    fluxOut[maxIndex:] = np.full(nw - maxIndex, fluxOut[minIndex])
    fluxOut = fluxOut - fluxOut[minIndex]

    return fluxOut



# Pre-processing class
import sys
from scipy.signal import medfilt
import numpy as np
from astrodash.preprocessing import ReadSpectrumFile, PreProcessSpectrum, ProcessingTools
from astrodash.array_tools import normalise_spectrum, zero_non_overlap_part


def limit_wavelength_range(wave, flux, minWave, maxWave):
    minIdx = (np.abs(wave - minWave)).argmin()
    maxIdx = (np.abs(wave - maxWave)).argmin()

    flux[:minIdx] = np.zeros(minIdx)
    flux[maxIdx:] = np.zeros(len(flux) - maxIdx)

    return flux


class PreProcessing(object):
    """ Pre-processes spectra before training """

    def __init__(self, filename, w0, w1, nw):
        self.filename = filename
        self.w0 = w0
        self.w1 = w1
        self.nw = nw
        self.numSplinePoints = 13
        self.processingTools = ProcessingTools()
        self.readSpectrumFile = ReadSpectrumFile(filename, w0, w1, nw)
        self.preProcess = PreProcessSpectrum(w0, w1, nw)

        self.spectrum = self.readSpectrumFile.file_extension()
        if len(self.spectrum) == 3:
            self.redshiftFromFile = True
        else:
            self.redshiftFromFile = False

    def two_column_data(self, z, smooth, minWave, maxWave):
        if self.redshiftFromFile is True:
            self.wave, self.flux, z = self.spectrum
        else:
            self.wave, self.flux = self.spectrum
        self.flux = normalise_spectrum(self.flux)
        self.flux = limit_wavelength_range(self.wave, self.flux, minWave, maxWave)
        self.wDensity = (self.w1 - self.w0) / self.nw  # Average wavelength spacing
        wavelengthDensity = (max(self.wave) - min(self.wave)) / len(self.wave)
        filterSize = int(self.wDensity / wavelengthDensity * smooth / 2) * 2 + 1
        preFiltered = medfilt(self.flux, kernel_size=filterSize)
        wave, deredshifted = self.readSpectrumFile.two_col_input_spectrum(self.wave, preFiltered, z)
        if len(wave) < 2:
            sys.exit("The redshifted spectrum of file: {0} is out of the classification range between {1} to {2} "
                     "Angstroms. Please remove this file from classification or reduce the redshift before re-running "
                     "the program.".format(self.filename, self.w0, self.w1))

        binnedwave, binnedflux, minIndex, maxIndex = self.preProcess.log_wavelength(wave, deredshifted)
        newflux, continuum = self.preProcess.continuum_removal(binnedwave, binnedflux, self.numSplinePoints, minIndex,
                                                               maxIndex)
        meanzero = self.preProcess.mean_zero(newflux, minIndex, maxIndex)
        apodized = self.preProcess.apodize(meanzero, minIndex, maxIndex)

        # filterSize = smooth * 2 + 1
        medianFiltered = medfilt(apodized, kernel_size=1)  # filterSize)
        fluxNorm = normalise_spectrum(medianFiltered)
        fluxNorm = zero_non_overlap_part(fluxNorm, minIndex, maxIndex, outerVal=0.5)

        # # PAPER PLOTS
        # import matplotlib.pyplot as plt
        #
        # plt.figure(num=1, figsize=(10, 6))
        # plt.plot(self.wave, self.flux, label='Raw', linewidth=1.8)
        # plt.plot(self.wave, preFiltered, label='Filtered', linewidth=1.8)
        # plt.ylim(-8, 8)
        # plt.xlabel('Wavelength ($\mathrm{\AA}$)', fontsize=19)
        # plt.ylabel('Relative Flux', fontsize=19)
        # plt.legend(frameon=False, fontsize=15, loc=1)
        # plt.xlim(2500, 9000)
        # plt.tick_params(labelsize=16)
        # plt.tight_layout()
        # plt.axhline(0, color='black', linewidth=0.5)
        # plt.savefig('/Users/danmuth/OneDrive/Documents/DASH/Paper/Figures/Filtering.pdf')
        #
        # plt.figure(num=2, figsize=(10, 6))
        # plt.plot(wave, deredshifted, label='De-redshifted', linewidth=1.8)
        # plt.plot(binnedwave, binnedflux, label='Log-wavelength binned', linewidth=1.8)
        # plt.xlabel('Wavelength ($\mathrm{\AA}$)', fontsize=19)
        # plt.ylabel('Relative Flux', fontsize=19)
        # plt.legend(frameon=False, fontsize=15, loc=1)
        # plt.xlim(2500, 9000)
        # plt.tick_params(labelsize=16)
        # plt.tight_layout()
        # plt.axhline(0, color='black', linewidth=0.5)
        # plt.savefig('/Users/danmuth/OneDrive/Documents/DASH/Paper/Figures/Deredshifting.pdf')
        #
        # plt.figure(num=3, figsize=(10, 6))
        # plt.plot(binnedwave, binnedflux, label='Log-wavelength binned', linewidth=1.8)
        # plt.plot(binnedwave, continuum, label='Continuum', linewidth=1.8)
        # plt.plot(binnedwave, newflux, label='Continuum divided', linewidth=1.8)
        # plt.xlabel('Wavelength ($\mathrm{\AA}$)', fontsize=19)
        # plt.ylabel('Relative Flux', fontsize=19)
        # plt.legend(frameon=False, fontsize=15, loc=1)
        # plt.xlim(2500, 9000)
        # plt.tick_params(labelsize=16)
        # plt.tight_layout()
        # plt.axhline(0, color='black', linewidth=0.5)
        # plt.savefig('/Users/danmuth/OneDrive/Documents/DASH/Paper/Figures/Continuum.pdf')
        #
        # plt.figure(num=4, figsize=(10, 6))
        # plt.plot(binnedwave, meanzero, label='Continuum divided', linewidth=1.8)
        # plt.plot(binnedwave, apodized, label='Apodized', linewidth=1.8)
        # # fluxNorm = (apodized - min(apodized)) / (max(apodized) - min(apodized))
        # plt.plot(binnedwave, fluxNorm, label='Normalised', linewidth=1.8)
        # plt.xlabel('Wavelength ($\mathrm{\AA}$)', fontsize=19)
        # plt.ylabel('Relative Flux', fontsize=19)
        # plt.legend(frameon=False, fontsize=15, loc=1)
        # plt.xlim(2500, 9000)
        # plt.tick_params(labelsize=16)
        # plt.tight_layout()
        # plt.axhline(0, color='black', linewidth=0.5)
        # plt.savefig('/Users/danmuth/OneDrive/Documents/DASH/Paper/Figures/Apodize.pdf')
        #
        # plt.show()

        return binnedwave, fluxNorm, minIndex, maxIndex, z


if __name__ == '__main__':
    fData = '/Users/danmuth/PycharmProjects/astrodash/templates/OzDES_data/ATEL_9570_Run25/DES16C2ma_C2_combined_160926_v10_b00.dat'
    preData = PreProcessing(fData, 3000, 10000, 1024)
    waveData, fluxData, minIData, maxIData, z = preData.two_column_data(0.24, 5, 3000, 10000)



import numpy as np
from astrodash.create_arrays import *
from astrodash.array_tools import normalise_spectrum, zero_non_overlap_part


class InputSpectra(object):
    def __init__(self, filename, z, nTypes, minAge, maxAge, ageBinSize, w0, w1, nw, typeList, smooth, minWave, maxWave,
                 hostList, nHostTypes):
        self.filename = filename
        self.z = z
        self.w0 = w0
        self.w1 = w1
        self.nw = nw
        self.nTypes = nTypes
        self.minAge = minAge
        self.maxAge = maxAge
        self.ageBinSize = ageBinSize
        self.typeList = typeList
        self.ageBinning = AgeBinning(self.minAge, self.maxAge, self.ageBinSize)
        self.numOfAgeBins = self.ageBinning.age_bin(self.maxAge) + 1
        self.nLabels = self.nTypes * self.numOfAgeBins * nHostTypes
        self.createLabels = CreateLabels(self.nTypes, self.minAge, self.maxAge, self.ageBinSize, self.typeList,
                                         hostList, nHostTypes)
        self.fileType = 'fits or twocolumn etc.'  # Will use later on
        self.typeNamesList = self.createLabels.type_names_list()
        self.smooth = smooth
        self.minWave = minWave
        self.maxWave = maxWave

    def redshifting(self):
        images = np.empty((0, int(self.nw)), np.float16)  # Number of pixels
        labels = np.empty((0, self.nLabels), np.uint16)  # Number of labels (SN types)
        filenames = []
        typeNames = []
        redshifts = []
        minMaxIndexes = []
        readSpectra = ReadSpectra(self.w0, self.w1, self.nw, self.filename)

        # Undo it's previous redshift)
        wave, flux, minIndex, maxIndex, z = readSpectra.input_spectrum(self.z, self.smooth, self.minWave, self.maxWave)
        nonzeroflux = flux[minIndex:maxIndex + 1]
        newflux = normalise_spectrum(nonzeroflux)
        newflux2 = np.concatenate((flux[0:minIndex], newflux, flux[maxIndex + 1:]))
        images = np.append(images, np.array([newflux2]), axis=0)  # images.append(newflux2)
        filenames.append(str(self.filename) + "_" + str(-z))
        redshifts.append(-z)
        minMaxIndexes.append((minIndex, maxIndex))
        # # Add white noise to regions outside minIndex to maxIndex
        # noise = np.zeros(self.nw)
        # noise[0:minIndex] = np.random.uniform(0.0, 1.0, minIndex)
        # noise[maxIndex:] = np.random.uniform(0.0, 1.0, self.nw - maxIndex)
        #
        # augmentedFlux = flux + noise
        # augmentedFlux = normalise_spectrum(augmentedFlux)
        # augmentedFlux = zero_non_overlap_part(augmentedFlux, minIndex, maxIndex)

        inputImages = np.array(images)
        inputFilenames = np.array(filenames)
        inputRedshifts = np.array(redshifts)

        return inputImages, inputFilenames, inputRedshifts, self.typeNamesList, minMaxIndexes

    def saveArrays(self):
        inputImages, inputFilenames, inputRedshifts, minMaxIndex = self.redshifting()
        np.savez_compressed('input_data.npz', inputImages=inputImages, inputFilenames=inputFilenames,
                            inputRedshifts=inputRedshifts, typeNamesList=self.typeNamesList)
##
# sfTemplateLocation = '/home/dan/Desktop/SNClassifying_Pre-alpha/templates/superfit_templates/sne/'
##sfFilename = 'Ia/sn1981b.max.dat'
##
##filename = sfFilename
##
##with open('data_files/training_params.pickle') as f:
##    nTypes, w0, w1, nw, minAge, maxAge, ageBinSize = pickle.load(f)
##    
##minZ = 0
##maxZ = 0.5
##
##InputSpectra(filename, minZ, maxZ, nTypes, minAge, maxAge, ageBinSize, w0, w1, nw).saveArrays()



import os
import astrodash
import pandas as pd
import numpy as np
import astropy.io.fits as afits


def read_all_atels(atelTextFile):
    """ Read the text file containing all OzDES ATel
    Returns the names, redshifts, type, phase and notes from the ATel
    atels Index(['Name', 'RA (J2000)', 'Dec (J2000)', 'Discovery Date (UT) ',
       'Discovery Mag (r)', 'Spectrum Date (UT)', 'Redshift', 'Type', 'Phase',
       'Notes'],
      dtype='object')
    """

    atels = pd.read_csv(atelTextFile, delimiter='|')
    atels['Name'] = atels['Name'].str.strip()
    pd.to_numeric(atels['Redshift'])
    atels['Filename'] = ''

    return atels


def main(spectraDir, atelTextFile, saveMatchesFilename):

    # Store all the file paths to the objects in this directory
    allFilePaths = os.listdir(spectraDir)

    # Get filenames and corresponding redshifts
    atels = read_all_atels(atelTextFile)
    for filename in allFilePaths:
        name = filename.replace('.fits', '')
        i = np.where(atels['Name'] == name)[0]
        if len(i) > 1:
            print("Something wrong with:", name)
        atels.at[i, 'Filename'] = os.path.join(spectraDir, filename)

    # Check how many observed runs
    count = 0
    for filename in atels.Filename:
        hdulist = afits.open(filename)
        run = hdulist[3].header['SOURCEF'].split('/')[0]
        for i in range(3, len(hdulist)):
            try:
                run_new = hdulist[i].header['SOURCEF'].split('/')[0]
                if run_new != run:
                    count += 1
                    print(count, len(atels.Filename), filename, i, run, run_new)
                    break
            except KeyError:
                pass

    # Classify and print the best matches
    classification = astrodash.Classify(atels.Filename.values, atels.Redshift.values, classifyHost=False, rlapScores=True, smooth=6)
    bestFits, redshifts, bestTypes, rlapFlag, matchesFlag, redshiftErrs = classification.list_best_matches(n=5, saveFilename=saveMatchesFilename)

    print("{0:17} | {1:5} | {2:8} | {3:10} | {4:6} | {5:10} | {6:10}".format("Name", "  z  ", "DASH_Fit", "  Age ", "Prob.", "Flag", "Wiki Fit"))
    for i in range(len(atels)):
        print("{0:17} | {1:5} | {2:8} | {3:10} | {4:6} | {5:10} | {6:10}".format(os.path.basename(atels.Filename[i]), round(atels.Redshift[i], 3), bestTypes[i][0], bestTypes[i][1], bestTypes[i][2], matchesFlag[i].replace(' matches',''), "{} {}".format(atels['Type'][i], atels['Phase'][i])))

    # Plot one of the matches
    classification.plot_with_gui(indexToPlot=7)


if __name__=='__main__':
    main(spectraDir='/Users/danmuth/PycharmProjects/astrodash/templates/OzDES_data/180413/',
         atelTextFile='/Users/danmuth/PycharmProjects/astrodash/templates/OzDES_data/all_atels.txt',
         saveMatchesFilename='DASH_matches_all_atels.txt')




import os
import sys
import pickle
import numpy as np
from astrodash.download_data_files import download_all_files
from astrodash.restore_model import LoadInputSpectra, BestTypesListSingleRedshift, get_training_parameters, \
    classification_split
from astrodash.false_positive_rejection import RlapCalc, combined_prob
from astrodash.read_binned_templates import load_templates, get_templates
from astrodash.calculate_redshift import get_median_redshift
from astrodash.read_from_catalog import catalogDict

try:
    from PyQt5 import QtGui
    from astrodash.gui_main import MainApp
except ImportError:
    print("Warning: You will need to install 'PyQt5' if you want to use the graphical interface. "
          "Using the automatic library will continue to work.")


class Classify(object):
    def __init__(self, filenames=[], redshifts=[], smooth=6, minWave=3500, maxWave=10000, classifyHost=False,
                 knownZ=True, rlapScores=True, data_files='models_v06'):
        """ Takes a list of filenames and corresponding redshifts for supernovae.
        Files should contain a single spectrum, and redshifts should be a list of corresponding redshift floats
        """
        # download_all_files('v01')
        self.filenames = filenames
        self.redshifts = redshifts
        self.smooth = smooth
        self.minWave = minWave
        self.maxWave = maxWave
        self.classifyHost = classifyHost
        self.numSpectra = len(filenames)
        self.scriptDirectory = os.path.dirname(os.path.abspath(__file__))
        if knownZ and redshifts != []:
            self.knownZ = True
        else:
            self.knownZ = False
        if not self.redshifts:
            self.redshifts = [None] * len(filenames)
        self.rlapScores = rlapScores
        self.pars = get_training_parameters()
        self.nw, w0, w1 = self.pars['nw'], self.pars['w0'], self.pars['w1']
        self.dwlog = np.log(w1 / w0) / self.nw
        self.wave = w0 * np.exp(np.arange(0, self.nw) * self.dwlog)
        self.snTemplates, self.galTemplates = load_templates(
            os.path.join(self.scriptDirectory, data_files, 'models/sn_and_host_templates.npz'))

        if self.knownZ:
            if classifyHost:
                self.modelFilename = os.path.join(self.scriptDirectory, data_files,
                                                  "models/zeroZ_classifyHost/tensorflow_model.ckpt")
            else:
                self.modelFilename = os.path.join(self.scriptDirectory, data_files,
                                                  "models/zeroZ/tensorflow_model.ckpt")
        else:
            if self.classifyHost:
                raise ValueError("A model that classifies the host while simulatenously not knowing redshift does not "
                                 "exist currently. Please try one of the other 3 models or check back at a later "
                                 "date. Contact the author for support or further queries.")
                # self.modelFilename = os.path.join(self.scriptDirectory, data_files,
                #                                   "models/agnosticZ_classifyHost/tensorflow_model.ckpt")
            else:
                self.modelFilename = os.path.join(self.scriptDirectory, data_files,
                                                  "models/agnosticZ/tensorflow_model.ckpt")

    def _get_images(self, filename, redshift):
        if redshift in list(catalogDict.keys()):
            redshift = 0
        loadInputSpectra = LoadInputSpectra(filename, redshift, self.smooth, self.pars, self.minWave, self.maxWave,
                                            self.classifyHost)
        inputImage, inputRedshift, typeNamesList, nw, nBins, inputMinMaxIndex = loadInputSpectra.input_spectra()

        return inputImage, typeNamesList, nw, nBins, inputMinMaxIndex, inputRedshift

    def _input_spectra_info(self):
        inputImages = np.empty((0, int(self.nw)), np.float16)
        inputMinMaxIndexes = []
        for i in range(self.numSpectra):
            f = self.filenames[i]
            if self.knownZ:
                z = self.redshifts[i]
            else:
                z = 0
            inputImage, typeNamesList, nw, nBins, inputMinMaxIndex, inputRedshift = self._get_images(f, z)
            self.redshifts[i] = -inputRedshift[0]
            inputImages = np.append(inputImages, inputImage, axis=0)
            inputMinMaxIndexes.append(inputMinMaxIndex[0])
        bestTypesList = BestTypesListSingleRedshift(self.modelFilename, inputImages, typeNamesList, self.nw, nBins)
        bestTypes = bestTypesList.bestTypes
        softmaxes = bestTypesList.softmaxOrdered
        bestLabels = bestTypesList.idx

        return bestTypes, softmaxes, bestLabels, inputImages, inputMinMaxIndexes

    def list_best_matches(self, n=5, saveFilename='DASH_matches.txt'):
        """Returns a list of lists of the the top n best matches for each spectrum"""
        bestTypes, softmaxes, bestLabels, inputImages, inputMinMaxIndexes = self._input_spectra_info()
        bestMatchLists = []
        bestBroadTypes = []
        rlapLabels = []
        matchesReliableLabels = []
        redshifts = []
        redshiftErrs = []
        for specNum in range(self.numSpectra):
            bestMatchList = []
            for i in range(20):
                host, name, age = classification_split(bestTypes[specNum][i])
                if not self.knownZ:
                    redshift, _, redshiftErr = self.calc_redshift(inputImages[specNum], name, age, inputMinMaxIndexes[specNum])
                    redshifts.append(redshift)
                    redshiftErrs.append(redshiftErr)
                prob = softmaxes[specNum][i]
                bestMatchList.append((host, name, age, prob))
            bestMatchList = np.array(bestMatchList)
            bestMatchLists.append(bestMatchList[0:n])
            bestBroadType, matchesReliableFlag = self.best_broad_type(bestMatchList)
            bestBroadTypes.append(bestBroadType)
            rlapLabel, rlapWarningBool = self.rlap_warning_label(bestTypes[specNum][0], inputImages[specNum],
                                                                 inputMinMaxIndexes[specNum])

            rlapLabels.append(rlapLabel)
            if matchesReliableFlag:
                matchesReliableLabels.append("Reliable matches")
            else:
                matchesReliableLabels.append("Unreliable matches")

        bestMatchLists = np.array(bestMatchLists)

        if not redshifts:
            redshifts = self.redshifts
            redshiftErrs = [None] * len(self.redshifts)
        else:
            redshifts = np.array(redshifts)
            redshiftErrs = np.array(redshiftErrs)

        if saveFilename:
            self.save_best_matches(bestMatchLists, redshifts, bestBroadTypes, rlapLabels, matchesReliableLabels,
                                   saveFilename)

        return bestMatchLists, redshifts, bestBroadTypes, rlapLabels, matchesReliableLabels, redshiftErrs

    def best_broad_type(self, bestMatchList):
        host, prevName, bestAge, probTotal, reliableFlag = combined_prob(bestMatchList[0:10])

        return (prevName, bestAge, probTotal), reliableFlag

    def rlap_warning_label(self, bestType, inputImage, inputMinMaxIndex):
        host, name, age = classification_split(bestType)
        snInfos, snNames, hostInfos, hostNames = get_templates(name, age, host, self.snTemplates, self.galTemplates,
                                                               self.nw)
        if snInfos != []:
            if self.rlapScores:
                templateImages = snInfos[:, 1]
                templateMinMaxIndexes = list(zip(snInfos[:, 2], snInfos[:, 3]))
                rlapCalc = RlapCalc(inputImage, templateImages, snNames, self.wave, inputMinMaxIndex,
                                    templateMinMaxIndexes)
                rlap, rlapWarningBool = rlapCalc.rlap_label()
                if rlapWarningBool:
                    rlapLabel = "Low rlap: {0}".format(rlap)
                else:
                    rlapLabel = "Good rlap: {0}".format(rlap)
            else:
                rlapLabel = "No rlap"
                rlapWarningBool = "None"
        else:
            rlapLabel = "(NO_TEMPLATES)"
            rlapWarningBool = "None"

        # import matplotlib
        # matplotlib.use('TkAgg')
        # import matplotlib.pyplot as plt
        # plt.plot(inputImage)
        # plt.plot(templateImage)
        # plt.show()

        return rlapLabel, rlapWarningBool

    def calc_redshift(self, inputFlux, snName, snAge, inputMinMaxIndex):
        host = "No Host"
        snInfos, snNames, hostInfos, hostNames = get_templates(snName, snAge, host, self.snTemplates, self.galTemplates,
                                                               self.nw)
        numOfSubTemplates = len(snNames)
        templateNames = snNames
        templateFluxes = []
        templateMinMaxIndexes = []
        for i in range(numOfSubTemplates):
            templateFluxes.append(snInfos[i][1])
            templateMinMaxIndexes.append((snInfos[i][2], snInfos[i][3]))

        redshift, crossCorr, medianName, redshiftErr = get_median_redshift(inputFlux, templateFluxes, self.nw,
                                                                           self.dwlog, inputMinMaxIndex,
                                                                           templateMinMaxIndexes, templateNames,
                                                                           outerVal=0.5)
        print(redshift)
        if redshift is None:
            return 0, np.zeros(1024)

        return round(redshift, 4), crossCorr, round(redshiftErr, 4)

    def save_best_matches(self, bestFits, redshifts, bestTypes, rlapLabels, matchesReliableLabels,
                          saveFilename='DASH_matches.txt'):
        with open(saveFilename, 'w') as f:
            for i in range(len(self.filenames)):
                f.write("%s   z=%s     %s      %s     %s\n %s\n\n" % (
                    str(self.filenames[i]).split('/')[-1], redshifts[i], bestTypes[i], rlapLabels[i],
                    matchesReliableLabels[i], bestFits[i]))
        print("Finished classifying %d spectra!" % len(self.filenames))

    def plot_with_gui(self, indexToPlot=0):
        app = QtGui.QApplication(sys.argv)
        form = MainApp(inputFilename=self.filenames[indexToPlot])
        if not isinstance(self.filenames[indexToPlot], (list, np.ndarray)) and not hasattr(self.filenames[indexToPlot],
                                                                                           'read'):  # Not an array and not a file-handle
            form.lineEditInputFilename.setText(self.filenames[indexToPlot])
        form.checkBoxKnownZ.setChecked(self.knownZ)
        form.checkBoxClassifyHost.setChecked(self.classifyHost)
        form.lineEditKnownZ.setText(str(self.redshifts[indexToPlot]))
        form.lineEditSmooth.setText(str(self.smooth))
        form.classifyHost = self.classifyHost
        form.select_tensorflow_model()
        form.fit_spectra()
        form.show()
        app.exec_()

# # EXAMPLE USAGE:
# classification = Classify(filenames=['/Users/dmuthukrishna/Users/dmuthukrishna/DES16E1dic_E1_combined_161125_v10_b00.dat',
#                                      '/Users/dmuthukrishna/Users/dmuthukrishna/DES16E1dic_E1_combined_161125_v10_b00.dat'],
#                           redshifts=[0.34, 0.13])
# print(classification.list_best_matches())



import pickle
import numpy as np
import zipfile
import gzip
import os
import random
import copy
from collections import OrderedDict
from astrodash.create_arrays import AgeBinning, CreateLabels, ArrayTools, CreateArrays
from astrodash.helpers import temp_list

random.seed(42)


class CreateTrainingSet(object):

    def __init__(self, snidTemplateLocation, snidTempFileList, w0, w1, nw, nTypes, minAge, maxAge, ageBinSize, typeList,
                 minZ, maxZ, numOfRedshifts, galTemplateLocation, galTempFileList, hostTypes, nHostTypes,
                 trainFraction):
        self.snidTemplateLocation = snidTemplateLocation
        self.snidTempFileList = snidTempFileList
        self.galTemplateLocation = galTemplateLocation
        self.galTempFileList = galTempFileList
        self.w0 = w0
        self.w1 = w1
        self.nw = nw
        self.nTypes = nTypes
        self.minAge = minAge
        self.maxAge = maxAge
        self.ageBinSize = ageBinSize
        self.typeList = typeList
        self.trainFraction = trainFraction
        self.ageBinning = AgeBinning(self.minAge, self.maxAge, self.ageBinSize)
        self.numOfAgeBins = self.ageBinning.age_bin(self.maxAge - 0.1) + 1
        self.nLabels = self.nTypes * self.numOfAgeBins * nHostTypes
        self.createArrays = CreateArrays(w0, w1, nw, nTypes, minAge, maxAge, ageBinSize, typeList, minZ, maxZ,
                                         numOfRedshifts, hostTypes, nHostTypes)
        self.arrayTools = ArrayTools(self.nLabels, self.nw)

    def type_amounts(self, labels):
        counts = self.arrayTools.count_labels(labels)

        return counts

    def all_templates_to_arrays(self, snTempFileList, galTemplateLocation):
        """
        Parameters
        ----------
        snTempFileList : list or dictionary
        galTemplateLocation

        Returns
        -------
        """
        images, labels, filenames, typeNames = self.createArrays.combined_sn_gal_arrays_multiprocessing(
            self.snidTemplateLocation, snTempFileList, galTemplateLocation, self.galTempFileList)

        arraysShuf = self.arrayTools.shuffle_arrays(images=images, labels=labels, filenames=filenames,
                                                    typeNames=typeNames, memmapName='all')

        typeAmounts = self.type_amounts(labels)

        return arraysShuf, typeAmounts

    def train_test_split(self):
        """
        Split training set before creating arrays.
        Maybe should change this to include ages in train/test split instead of just SN files.
        """
        snTempFileList = copy.copy(self.snidTempFileList)
        fileList = temp_list(snTempFileList)
        snAndAgeIdxDict = OrderedDict()
        spectraList = []

        # SPLIT BY SPECTRA
        # Get number of spectra per file
        for i, sn in enumerate(fileList):
            with open(os.path.join(self.snidTemplateLocation, sn), 'r') as FileObj:
                for lineNum, line in enumerate(FileObj):
                    # Read Header Info
                    if lineNum == 0:
                        header = (line.strip('\n')).split(' ')
                        header = [x for x in header if x != '']
                        numAges, nwx, w0x, w1x, mostKnots, tname, dta, ttype, ittype, itstype = header
                        numAges, mostKnots = map(int, (numAges, mostKnots))
                    elif lineNum == mostKnots + 2:
                        ages = np.array(line.split()[1:]).astype(float)
                        agesIndexesInRange = np.where((ages >= self.minAge) & (ages <= self.maxAge))[0]
                        snAndAgeIdxDict[sn] = agesIndexesInRange
                        for ageIdx in agesIndexesInRange:
                            spectraList.append((sn, ageIdx))

        # Split train/test
        random.shuffle(spectraList)
        trainSize = int(self.trainFraction * len(spectraList))
        trainSpectra = spectraList[:trainSize]
        testSpectra = spectraList[trainSize:]

        trainDict, testDict = OrderedDict(), OrderedDict()
        for k, v in trainSpectra:
            trainDict.setdefault(k, []).append(v)
        for k, v in testSpectra:
            testDict.setdefault(k, []).append(v)

        # # SPLIT BY FILENAME INSTEAD OF BY SPECTRA
        # random.Random(42).shuffle(fileList)
        #
        # trainSize = int(self.trainFraction * len(fileList))
        # dirName = os.path.dirname(self.snidTempFileList)
        # trainListFileName = os.path.join(dirName, 'train_templist.txt')
        # testListFileName = os.path.join(dirName, 'test_templist.txt')
        #
        # # Save train set file list
        # with open(trainListFileName, 'w') as f:
        #     for line in fileList[:trainSize]:
        #         f.write("%s\n" % line)
        #
        # # Save test set file list
        # with open(testListFileName, 'w') as f:
        #     for line in fileList[trainSize:]:
        #         f.write("%s\n" % line)
        print("trainDict", trainDict)
        print("testDict", testDict)
        return trainDict, testDict

    def sort_data(self):
        if self.trainFraction == 1.0:
            arrays, typeAmounts = self.all_templates_to_arrays(self.snidTempFileList, self.galTemplateLocation)
            trainImages, trainLabels, trainFilenames, trainTypeNames = arrays['images'], arrays['labels'], arrays[
                'filenames'], arrays['typeNames']
            testImages, testLabels, testFilenames, testTypeNames = trainImages[-1:], trainLabels[-1:], trainFilenames[
                                                                                                       -1:], trainTypeNames[
                                                                                                             -1:]
        else:
            trainDict, testDict = self.train_test_split()

            arraysTrain, typeAmountsTrain = self.all_templates_to_arrays(trainDict, self.galTemplateLocation)
            trainImages, trainLabels, trainFilenames, trainTypeNames = arraysTrain['images'], arraysTrain['labels'], \
                                                                       arraysTrain['filenames'], arraysTrain[
                                                                           'typeNames']

            arraysTest, typeAmountsTest = self.all_templates_to_arrays(testDict, None)
            testImages, testLabels, testFilenames, testTypeNames = arraysTest['images'], arraysTest['labels'], \
                                                                   arraysTest['filenames'], arraysTest['typeNames']

        # trainPercentage = self.trainFraction
        # testPercentage = 1.0 - self.trainFraction
        #
        # arrays, typeAmounts = self.all_templates_to_arrays(self.snidTempFileList, self.galTemplateLocation)
        # images, labels, filenames, typeNames = arrays['images'], arrays['labels'], arrays['filenames'], arrays['typeNames']
        #
        # trainSize = int(trainPercentage * len(images))
        # testSize = int(testPercentage * len(images))
        #
        # trainImages = images[:trainSize]
        # testImages = images[trainSize: trainSize + testSize]
        # trainLabels = labels[:trainSize]
        # testLabels = labels[trainSize: trainSize + testSize]
        # trainFilenames = filenames[:trainSize]
        # testFilenames = filenames[trainSize: trainSize + testSize]
        # trainTypeNames = typeNames[:trainSize]
        # testTypeNames = typeNames[trainSize: trainSize + testSize]

        typeAmounts = self.type_amounts(trainLabels)

        return ((trainImages, trainLabels, trainFilenames, trainTypeNames),
                (testImages, testLabels, testFilenames, testTypeNames),
                typeAmounts)


class SaveTrainingSet(object):
    def __init__(self, snidTemplateLocation, snidTempFileList, w0, w1, nw, nTypes, minAge, maxAge, ageBinSize, typeList,
                 minZ, maxZ, numOfRedshifts, galTemplateLocation=None, galTempFileList=None, hostTypes=None,
                 nHostTypes=1, trainFraction=0.8):
        self.snidTemplateLocation = snidTemplateLocation
        self.snidTempFileList = snidTempFileList
        self.w0 = w0
        self.w1 = w1
        self.nw = nw
        self.nTypes = nTypes
        self.minAge = minAge
        self.maxAge = maxAge
        self.ageBinSize = ageBinSize
        self.typeList = typeList
        self.createLabels = CreateLabels(nTypes, minAge, maxAge, ageBinSize, typeList, hostTypes, nHostTypes)

        self.createTrainingSet = CreateTrainingSet(snidTemplateLocation, snidTempFileList, w0, w1, nw, nTypes, minAge,
                                                   maxAge, ageBinSize, typeList, minZ, maxZ, numOfRedshifts,
                                                   galTemplateLocation, galTempFileList, hostTypes, nHostTypes,
                                                   trainFraction)
        self.sortData = self.createTrainingSet.sort_data()
        self.trainImages = self.sortData[0][0]
        self.trainLabels = self.sortData[0][1]
        self.trainFilenames = self.sortData[0][2]
        self.trainTypeNames = self.sortData[0][3]
        self.testImages = self.sortData[1][0]
        self.testLabels = self.sortData[1][1]
        self.testFilenames = self.sortData[1][2]
        self.testTypeNames = self.sortData[1][3]
        self.typeAmounts = self.sortData[2]

        self.typeNamesList = self.createLabels.type_names_list()

    def type_amounts(self):
        for i in range(len(self.typeNamesList)):
            print(str(self.typeAmounts[i]) + ": " + str(self.typeNamesList[i]))
        return self.typeNamesList, self.typeAmounts

    def save_arrays(self, saveFilename):
        arraysToSave = {'trainImages.npy.gz': self.trainImages, 'trainLabels.npy.gz': self.trainLabels,
                        'testImages.npy.gz': self.testImages, 'testLabels.npy.gz': self.testLabels,
                        'testTypeNames.npy.gz': self.testTypeNames, 'typeNamesList.npy.gz': self.typeNamesList,
                        'trainFilenames.npy.gz': self.trainFilenames, 'trainTypeNames.npy.gz': self.trainTypeNames}

        try:
            print("SIZE OF ARRAYS TRAINING:")
            print(self.trainImages.nbytes, self.testImages.nbytes)
            print(self.trainLabels.nbytes, self.testLabels.nbytes)
            print(self.trainFilenames.nbytes, self.testFilenames.nbytes)
            print(self.trainTypeNames.nbytes, self.testTypeNames.nbytes)
        except:
            print("Exception Raised")

        for filename, array in arraysToSave.items():
            f = gzip.GzipFile(filename, "w")
            np.save(file=f, arr=array)
            f.close()

        with zipfile.ZipFile(saveFilename, 'w') as myzip:
            for f in arraysToSave.keys():
                myzip.write(f)

        print("Saved Training Set to: " + saveFilename)

        # Delete npy.gz files
        for filename in arraysToSave.keys():
            os.remove(filename)


def create_training_set_files(dataDirName, minZ=0, maxZ=0, numOfRedshifts=80, trainWithHost=True, classifyHost=False,
                              trainFraction=0.8):
    with open(os.path.join(dataDirName, 'training_params.pickle'), 'rb') as f1:
        pars = pickle.load(f1)
    nTypes, w0, w1, nw, minAge, maxAge, ageBinSize, typeList = pars['nTypes'], pars['w0'], pars['w1'], \
                                                               pars['nw'], pars['minAge'], pars['maxAge'], \
                                                               pars['ageBinSize'], pars['typeList']
    hostList, nHostTypes = None, 1

    scriptDirectory = os.path.dirname(os.path.abspath(__file__))

    snidTemplateLocation = os.path.join(scriptDirectory, "../templates/training_set/")
    snidTempFileList = snidTemplateLocation + 'templist.txt'
    if trainWithHost:
        galTemplateLocation = os.path.join(scriptDirectory, "../templates/superfit_templates/gal/")
        galTempFileList = galTemplateLocation + 'gal.list'
        if classifyHost:
            hostList = pars['galTypeList']
            nHostTypes = len(hostList)
    else:
        galTemplateLocation, galTempFileList = None, None

    saveTrainingSet = SaveTrainingSet(snidTemplateLocation, snidTempFileList, w0, w1, nw, nTypes, minAge, maxAge,
                                      ageBinSize, typeList, minZ, maxZ, numOfRedshifts, galTemplateLocation,
                                      galTempFileList, hostList, nHostTypes, trainFraction)
    typeNamesList, typeAmounts = saveTrainingSet.type_amounts()

    saveFilename = os.path.join(dataDirName, 'training_set.zip')
    saveTrainingSet.save_arrays(saveFilename)

    return saveFilename


if __name__ == '__main__':
    trainingSetFilename = create_training_set_files('data_files/', minZ=0, maxZ=0, numOfRedshifts=80,
                                                    trainWithHost=False, classifyHost=False, trainFraction=0.8)

# # Split by filename instead of by spectra
# snTempFileList = copy.copy(self.snidTempFileList)
# fileList = temp_list(snTempFileList)
# random.Random(42).shuffle(fileList)
#
# trainSize = int(self.trainFraction * len(fileList))
# dirName = os.path.dirname(self.snidTempFileList)
# trainListFileName = os.path.join(dirName, 'train_templist.txt')
# testListFileName = os.path.join(dirName, 'test_templist.txt')
#
# # Save train set file list
# with open(trainListFileName, 'w') as f:
#     for line in fileList[:trainSize]:
#         f.write("%s\n" % line)
#
# # Save test set file list
# with open(testListFileName, 'w') as f:
#     for line in fileList[trainSize:]:
#         f.write("%s\n" % line)



import numpy as np
from random import shuffle
import multiprocessing as mp
import itertools

from astrodash.helpers import temp_list, div0
from astrodash.sn_processing import PreProcessing
from astrodash.combine_sn_and_host import training_template_data
from astrodash.preprocessing import ProcessingTools
from astrodash.array_tools import zero_non_overlap_part, normalise_spectrum

try:
    from imblearn import over_sampling
    IMBLEARN_EXISTS = True
except ImportError:
    IMBLEARN_EXISTS = False


class AgeBinning(object):
    def __init__(self, minAge, maxAge, ageBinSize):
        self.minAge = minAge
        self.maxAge = maxAge
        self.ageBinSize = ageBinSize

    def age_bin(self, age):
        ageBin = int(round(age / self.ageBinSize)) - int(round(self.minAge / self.ageBinSize))

        return ageBin

    def age_labels(self):
        ageLabels = []

        ageBinPrev = 0
        ageLabelMin = self.minAge
        for age in np.arange(self.minAge, self.maxAge, 0.5):
            ageBin = self.age_bin(age)

            if ageBin != ageBinPrev:
                ageLabelMax = int(round(age))
                ageLabels.append(str(int(ageLabelMin)) + " to " + str(ageLabelMax))
                ageLabelMin = ageLabelMax

            ageBinPrev = ageBin

        ageLabels.append(str(int(ageLabelMin)) + " to " + str(int(self.maxAge)))

        return ageLabels


class CreateLabels(object):

    def __init__(self, nTypes, minAge, maxAge, ageBinSize, typeList, hostList, nHostTypes):
        self.nTypes = nTypes
        self.minAge = minAge
        self.maxAge = maxAge
        self.ageBinSize = ageBinSize
        self.typeList = typeList
        self.ageBinning = AgeBinning(self.minAge, self.maxAge, self.ageBinSize)
        self.numOfAgeBins = self.ageBinning.age_bin(self.maxAge - 0.1) + 1
        self.nLabels = self.nTypes * self.numOfAgeBins
        self.ageLabels = self.ageBinning.age_labels()
        self.hostList = hostList
        self.nHostTypes = nHostTypes

    def label_array(self, ttype, age, host=None):
        ageBin = self.ageBinning.age_bin(age)

        try:
            typeIndex = self.typeList.index(ttype)
        except ValueError as err:
            raise Exception("INVALID TYPE: {0}".format(err))

        if host is None:
            labelArray = np.zeros((self.nTypes, self.numOfAgeBins))
            labelArray[typeIndex][ageBin] = 1
            labelArray = labelArray.flatten()
            typeName = ttype + ": " + self.ageLabels[ageBin]
        else:
            hostIndex = self.hostList.index(host)
            labelArray = np.zeros((self.nHostTypes, self.nTypes, self.numOfAgeBins))
            labelArray[hostIndex][typeIndex][ageBin] = 1
            labelArray = labelArray.flatten()
            typeName = "{}: {}: {}".format(host, ttype, self.ageLabels[ageBin])

        labelIndex = np.argmax(labelArray)

        return labelIndex, typeName

    def type_names_list(self):
        typeNamesList = []
        if self.hostList is None:
            for tType in self.typeList:
                for ageLabel in self.ageBinning.age_labels():
                    typeNamesList.append("{}: {}".format(tType, ageLabel))
        else:
            for host in self.hostList:
                for tType in self.typeList:
                    for ageLabel in self.ageBinning.age_labels():
                        typeNamesList.append("{}: {}: {}".format(host, tType, ageLabel))

        return np.array(typeNamesList)


class ReadSpectra(object):

    def __init__(self, w0, w1, nw, snFilename, galFilename=None):
        self.w0 = w0
        self.w1 = w1
        self.nw = nw
        self.snFilename = snFilename
        self.galFilename = galFilename
        if galFilename is None:
            self.data = PreProcessing(snFilename, w0, w1, nw)

    def sn_plus_gal_template(self, snAgeIdx, snCoeff, galCoeff, z):
        wave, flux, minIndex, maxIndex, nCols, ages, tType = training_template_data(snAgeIdx, snCoeff, galCoeff, z,
                                                                                    self.snFilename, self.galFilename,
                                                                                    self.w0, self.w1, self.nw)

        return wave, flux, nCols, ages, tType, minIndex, maxIndex

    def input_spectrum(self, z, smooth, minWave, maxWave):
        wave, flux, minIndex, maxIndex, z = self.data.two_column_data(z, smooth, minWave, maxWave)

        return wave, flux, int(minIndex), int(maxIndex), z


class ArrayTools(object):

    def __init__(self, nLabels, nw):
        self.nLabels = nLabels
        self.nw = nw

    def shuffle_arrays(self, memmapName='', **kwargs):
        """ Must take images and labels as arguments with the keyword specified.
        Can optionally take filenames and typeNames as arguments """
        arraySize = len(kwargs['labels'])
        if arraySize == 0:
            return kwargs

        kwargShuf = {}
        self.randnum = np.random.randint(10000)
        for key in kwargs:
            if key == 'images':
                arrayShuf = np.memmap('shuffled_{}_{}_{}.dat'.format(key, memmapName, self.randnum), dtype=np.float16,
                                      mode='w+', shape=(arraySize, int(self.nw)))
            elif key == 'labels':
                arrayShuf = np.memmap('shuffled_{}_{}_{}.dat'.format(key, memmapName, self.randnum), dtype=np.uint16,
                                      mode='w+', shape=arraySize)
            else:
                arrayShuf = np.memmap('shuffled_{}_{}_{}.dat'.format(key, memmapName, self.randnum), dtype=object,
                                      mode='w+', shape=arraySize)
            kwargShuf[key] = arrayShuf

        print("Shuffling...")
        # Randomise order
        p = np.random.permutation(len(kwargs['labels']))
        for key in kwargs:
            assert len(kwargs[key]) == arraySize
            print(key, "shuffling...")
            print(len(p))
            kwargShuf[key] = kwargs[key][p]

        return kwargShuf

    def count_labels(self, labels):
        counts = np.zeros(self.nLabels)

        for i in range(len(labels)):
            counts[labels[i]] += 1

        return counts

    def augment_data(self, flux, stdDevMean=0.05, stdDevStdDev=0.05):
        minIndex, maxIndex = ProcessingTools().min_max_index(flux, outerVal=0.5)
        noise = np.zeros(self.nw)
        stdDev = abs(np.random.normal(stdDevMean, stdDevStdDev))  # randomised standard deviation
        noise[minIndex:maxIndex] = np.random.normal(0, stdDev, maxIndex - minIndex)
        # # Add white noise to regions outside minIndex to maxIndex
        # noise[0:minIndex] = np.random.uniform(0.0, 1.0, minIndex)
        # noise[maxIndex:] = np.random.uniform(0.0, 1.0, self.nw-maxIndex)

        augmentedFlux = flux + noise
        augmentedFlux = normalise_spectrum(augmentedFlux)
        augmentedFlux = zero_non_overlap_part(augmentedFlux, minIndex, maxIndex, outerVal=0.5)

        return augmentedFlux


class OverSampling(ArrayTools):
    def __init__(self, nLabels, nw, **kwargs):
        """ Must take images and labels as arguments with the keyword specified.
        Can optionally take filenames and typeNames as arguments """
        ArrayTools.__init__(self, nLabels, nw)
        self.kwargs = kwargs

        counts = self.count_labels(self.kwargs['labels'])
        print("Before OverSample")  #
        print(counts)  #

        self.overSampleAmount = np.rint(div0(1 * max(counts), counts))  # ignore zeros in counts
        self.overSampleArraySize = int(sum(np.array(self.overSampleAmount, int) * counts))
        print(np.array(self.overSampleAmount, int) * counts)
        print(np.array(self.overSampleAmount, int))
        print(self.overSampleArraySize, len(self.kwargs['labels']))
        self.kwargOverSampled = {}
        self.randnum = np.random.randint(10000)
        for key in self.kwargs:
            if key == 'images':
                arrayOverSampled = np.memmap('oversampled_{}_{}.dat'.format(key, self.randnum), dtype=np.float16,
                                             mode='w+',
                                             shape=(self.overSampleArraySize, int(self.nw)))
            elif key == 'labels':
                arrayOverSampled = np.memmap('oversampled_{}_{}.dat'.format(key, self.randnum), dtype=np.uint16,
                                             mode='w+',
                                             shape=self.overSampleArraySize)
            else:
                arrayOverSampled = np.memmap('oversampled_{}_{}.dat'.format(key, self.randnum), dtype=object, mode='w+',
                                             shape=self.overSampleArraySize)
            self.kwargOverSampled[key] = arrayOverSampled

        self.kwargShuf = self.shuffle_arrays(memmapName='pre-oversample_{}'.format(self.randnum), **self.kwargs)
        print(len(self.kwargShuf['labels']))

    def oversample_mp(self, i_in, offset_in, std_in, labelIndex_in):
        print('oversampling', i_in, len(self.kwargShuf['labels']))
        oversampled = {key: [] for key in self.kwargs}
        repeatAmount = int(self.overSampleAmount[labelIndex_in])
        for r in range(repeatAmount):
            for key in self.kwargs:
                if key == 'images':
                    oversampled[key].append(
                        self.augment_data(self.kwargShuf[key][i_in], stdDevMean=0.05, stdDevStdDev=std_in))
                else:
                    oversampled[key].append(self.kwargShuf[key][i_in])
        return oversampled, offset_in, repeatAmount

    def collect_results(self, result):
        """Uses apply_async's callback to setup up a separate Queue for each process"""
        oversampled_in, offset_in, repeatAmount = result
        for key in self.kwargs:
            rlength_array = np.array(oversampled_in[key])
            self.kwargOverSampled[key][offset_in:repeatAmount + offset_in] = rlength_array[:]

    def over_sample_arrays(self, smote=False):
        if smote:
            return self.smote_oversample()
        else:
            return self.minority_oversample_with_noise()

    def minority_oversample_with_noise(self):
        offset = 0
        # pool = mp.Pool()
        for i in range(len(self.kwargShuf['labels'])):
            labelIndex = self.kwargShuf['labels'][i]
            if self.overSampleAmount[labelIndex] < 10:
                std = 0.03
            else:
                std = 0.05
            # pool.apply_async(self.oversample_mp, args=(i, offset, std, labelIndex), callback=self.collect_results)
            self.collect_results(self.oversample_mp(i, offset, std, labelIndex))
            offset += int(self.overSampleAmount[labelIndex])
        # pool.close()
        # pool.join()

        # for i, output in enumerate(outputs):
        #     self.collect_results(output)
        #     print('combining results...', i, len(outputs))

        print("Before Shuffling")
        self.kwargOverSampledShuf = self.shuffle_arrays(memmapName='oversampled_{}'.format(self.randnum),
                                                        **self.kwargOverSampled)
        print("After Shuffling")

        return self.kwargOverSampledShuf

    def smote_oversample(self):
        sm = over_sampling.SMOTE(random_state=42, n_jobs=30)
        images, labels = sm.fit_sample(X=self.kwargShuf['images'], y=self.kwargShuf['labels'])

        self.kwargOverSampledShuf = self.shuffle_arrays(memmapName='oversampled_smote_{}'.format(self.randnum),
                                                        images=images, labels=labels)

        return self.kwargOverSampledShuf


class CreateArrays(object):
    def __init__(self, w0, w1, nw, nTypes, minAge, maxAge, ageBinSize, typeList, minZ, maxZ, numOfRedshifts,
                 hostTypes=None, nHostTypes=None):
        self.w0 = w0
        self.w1 = w1
        self.nw = nw
        self.nTypes = nTypes
        self.minAge = minAge
        self.maxAge = maxAge
        self.ageBinSize = ageBinSize
        self.typeList = typeList
        self.minZ = minZ
        self.maxZ = maxZ
        self.numOfRedshifts = numOfRedshifts
        self.ageBinning = AgeBinning(minAge, maxAge, ageBinSize)
        self.numOfAgeBins = self.ageBinning.age_bin(maxAge - 0.1) + 1
        self.nLabels = nTypes * self.numOfAgeBins * nHostTypes
        self.createLabels = CreateLabels(self.nTypes, self.minAge, self.maxAge, self.ageBinSize, self.typeList,
                                         hostTypes, nHostTypes)
        self.hostTypes = hostTypes

    def combined_sn_gal_templates_to_arrays(self, args):
        snTemplateLocation, snTempList, galTemplateLocation, galTempList, snFractions, ageIndexes = args
        images = np.empty((0, int(self.nw)), np.float16)  # Number of pixels
        labelsIndexes = []
        filenames = []
        typeNames = []

        for j, gal in enumerate(galTempList):
            galFilename = galTemplateLocation + gal if galTemplateLocation is not None else None
            for i, sn in enumerate(snTempList):
                nCols = 15
                readSpectra = ReadSpectra(self.w0, self.w1, self.nw, snTemplateLocation + sn, galFilename)
                for ageidx in ageIndexes[sn]:
                    if ageidx >= nCols:
                        break
                    for snCoeff in snFractions:
                        galCoeff = 1 - snCoeff
                        if self.numOfRedshifts == 1:
                            redshifts = [self.minZ]
                        else:
                            redshifts = np.random.uniform(low=self.minZ, high=self.maxZ, size=self.numOfRedshifts)
                        for z in redshifts:
                            tempWave, tempFlux, nCols, ages, tType, tMinIndex, tMaxIndex = readSpectra.sn_plus_gal_template(
                                ageidx, snCoeff, galCoeff, z)
                            if tMinIndex == tMaxIndex or not tempFlux.any():
                                print("NO DATA for {} {} ageIdx:{} z>={}".format(galTempList[j], snTempList[i], ageidx,
                                                                                 z))
                                break

                            if self.minAge < float(ages[ageidx]) < self.maxAge:
                                if self.hostTypes is None:  # Checks if we are classifying by host as well
                                    labelIndex, typeName = self.createLabels.label_array(tType, ages[ageidx], host=None)
                                else:
                                    labelIndex, typeName = self.createLabels.label_array(tType, ages[ageidx],
                                                                                         host=galTempList[j])
                                if tMinIndex > (self.nw - 1):
                                    continue
                                nonzeroflux = tempFlux[tMinIndex:tMaxIndex + 1]
                                newflux = (nonzeroflux - min(nonzeroflux)) / (max(nonzeroflux) - min(nonzeroflux))
                                newflux2 = np.concatenate((tempFlux[0:tMinIndex], newflux, tempFlux[tMaxIndex + 1:]))
                                images = np.append(images, np.array([newflux2]), axis=0)
                                labelsIndexes.append(
                                    labelIndex)  # labels = np.append(labels, np.array([label]), axis=0)
                                filenames.append(
                                    "{0}_{1}_{2}_{3}_snCoeff{4}_z{5}".format(snTempList[i], tType, str(ages[ageidx]),
                                                                             galTempList[j], snCoeff, (z)))
                                typeNames.append(typeName)
                print(snTempList[i], nCols, galTempList[j])

        return images, np.array(labelsIndexes).astype(int), np.array(filenames), np.array(typeNames)

    def collect_results(self, result):
        """Uses apply_async's callback to setup up a separate Queue for each process"""
        imagesPart, labelsPart, filenamesPart, typeNamesPart = result
        self.images.extend(imagesPart)
        self.labelsIndexes.extend(labelsPart)
        self.filenames.extend(filenamesPart)
        self.typeNames.extend(typeNamesPart)

    def combined_sn_gal_arrays_multiprocessing(self, snTemplateLocation, snTempFileList, galTemplateLocation,
                                               galTempFileList):
        # TODO: Maybe do memory mapping for these arrays
        self.images = []
        self.labelsIndexes = []
        self.filenames = []
        self.typeNames = []

        if galTemplateLocation is None or galTempFileList is None:
            galTempList = [None]
            galTemplateLocation = None
            snFractions = [1.0]
        else:
            galTempList = temp_list(galTempFileList)
            snFractions = [0.99, 0.98, 0.95, 0.93, 0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1]

        if isinstance(snTempFileList, dict):
            snTempList = list(snTempFileList.keys())
            ageIndexesDict = snTempFileList
        else:
            snTempList = temp_list(snTempFileList)
            ageIndexesDict = None

        galAndSnTemps = list(itertools.product(galTempList, snTempList))
        argsList = []
        for gal, sn in galAndSnTemps:
            if ageIndexesDict is not None:
                ageIdxDict = {k: ageIndexesDict[k] for k in (sn,)}
            else:
                ageIdxDict = {k: range(0, 1000) for k in (sn,)}
            argsList.append((snTemplateLocation, [sn], galTemplateLocation, [gal], snFractions, ageIdxDict))

        pool = mp.Pool()
        results = pool.map_async(self.combined_sn_gal_templates_to_arrays, argsList)
        pool.close()
        pool.join()

        outputs = results.get()
        for i, output in enumerate(outputs):
            self.collect_results(output)
            print('combining results...', i, len(outputs))

        self.images = np.array(self.images)
        self.labelsIndexes = np.array(self.labelsIndexes)
        self.filenames = np.array(self.filenames)
        self.typeNames = np.array(self.typeNames)

        print("Completed Creating Arrays!")

        return self.images, self.labelsIndexes.astype(np.uint16), self.filenames, self.typeNames



import os
import sys
import shutil
from astrodash.unzip_data_files import unzip_data_files


def download_file(filename, urlpath, printStatus, scriptDirectory):
    dataFilename = os.path.join(scriptDirectory, filename)
    if (not os.path.isfile(dataFilename)) and (not os.path.isdir(dataFilename.strip(".zip"))):
        print(printStatus)
        if sys.version_info[0] < 3:
            import urllib
            urllib.urlretrieve(urlpath, dataFilename)
        else:
            import urllib.request
            urllib.request.urlretrieve(urlpath, dataFilename)

        print("Download complete! File saved to %s." % dataFilename)
        if filename.endswith(".zip"):
            unzip_data_files(filename)


def delete_previous_versions(oldFilenames, scriptDirectory):
    for oldFilename in oldFilenames:
        dataFilename = os.path.join(scriptDirectory, oldFilename)
        if os.path.isfile(dataFilename):
            print("Deleting previous version of data file: %s" % oldFilename)
            os.remove(dataFilename)
        if os.path.isdir(dataFilename.strip('.zip')):
            shutil.rmtree(dataFilename.strip('.zip'))


def download_all_files(zipVersion):
    scriptDirectory = os.path.dirname(os.path.abspath(__file__))

    oldFilenames = ['models_v01.zip', 'models_v02.zip', 'models_v03.zip', 'models_v04.zip', 'models_v05.zip']

    delete_previous_versions(oldFilenames, scriptDirectory)

    saveFilenames = ['models_{0}.zip'.format(zipVersion)]

    urlpaths = ["https://github.com/daniel-muthukrishna/astrodash/blob/master/astrodash/models_{0}.zip?raw=true".format(
        zipVersion)]

    printStatuses = [
        "Downloading data files from {0}... \nThis download contains the Tensorflow models trained using deep learning. \n"
        "The file to be downloaded is ~200MB. However, this is a one time download, and will only occur\n"
        "the very first time that this version of DASH is installed.".format(urlpaths[0])]

    for i in range(len(urlpaths)):
        download_file(saveFilenames[i], urlpaths[i], printStatuses[i], scriptDirectory)


if __name__ == '__main__':
    download_all_files('v06')



import os
import pickle
import matplotlib.pyplot as plt
import itertools
import numpy as np
import tensorflow as tf
from astrodash.multilayer_convnet import convnet_variables


def plot_confusion_matrix(cm, classes, normalize=False, title='Confusion matrix', cmap=plt.cm.RdBu, fig_dir='.',
                          name='', fontsize_labels=15, fontsize_matrix=18):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    """

    np.savetxt(os.path.join(fig_dir, 'confusion_matrix_raw_%s.csv' % name), cm)

    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        print("Normalized confusion matrix")
    else:
        print('Confusion matrix, without normalization')

    print(cm)
    # Multiply off diagonal by -1
    off_diag = ~np.eye(cm.shape[0], dtype=bool)
    cm[off_diag] *= -1
    np.savetxt(os.path.join(fig_dir, 'confusion_matrix_%s.csv' % name), cm)
    print(cm)

    fig = plt.figure(figsize=(15, 12))
    plt.imshow(cm, interpolation='nearest', cmap=cmap, vmin=-1, vmax=1)
    plt.title(title)
    cb = plt.colorbar()
    cb.ax.set_yticklabels(cb.ax.get_yticklabels(), fontsize=16)
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=90, fontsize=fontsize_labels)
    plt.yticks(tick_marks, classes, fontsize=fontsize_labels)

    fmt = '.2f' if normalize else 'd'
    thresh = cm.max() / 2.
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(j, i, format(abs(cm[i, j]), fmt), horizontalalignment="center",
                 color="white" if abs(cm[i, j]) > thresh else "black", fontsize=fontsize_matrix)

    plt.tight_layout()
    plt.ylabel('True label', fontsize=18)
    plt.xlabel('Predicted label', fontsize=18)
    plt.tight_layout()
    plt.savefig(os.path.join(fig_dir, 'confusion_matrix_%s.pdf' % name))


def get_aggregated_conf_matrix(aggregateIndexes, testLabels, predictedLabels):
    testLabelsAggregated = np.digitize(testLabels, aggregateIndexes) - 1
    predictedLabelsAggregated = np.digitize(predictedLabels, aggregateIndexes) - 1
    confMatrixAggregated = tf.confusion_matrix(testLabelsAggregated, predictedLabelsAggregated).eval()
    np.set_printoptions(precision=2)
    print(confMatrixAggregated)

    return confMatrixAggregated


def calc_model_metrics(modelFilename, testLabels, testImages, testTypeNames, typeNamesList, snTypes=None, fig_dir='.'):
    tf.reset_default_graph()
    nw = len(testImages[0])
    nBins = len(typeNamesList)
    imWidthReduc = 8
    imWidth = 32  # Image size and width

    x, y_, keep_prob, y_conv, W, b = convnet_variables(imWidth, imWidthReduc, nw, nBins)

    saver = tf.train.Saver()

    with tf.Session() as sess:
        saver.restore(sess, modelFilename)

        yy = y_conv.eval(feed_dict={x: testImages, keep_prob: 1.0})

        # CONFUSION MATRIX
        predictedLabels = []
        for i, name in enumerate(testTypeNames):
            predictedLabels.append(np.argmax(yy[i]))
        predictedLabels = np.array(predictedLabels)
        confMatrix = tf.confusion_matrix(testLabels, predictedLabels).eval()

        # Aggregate age conf matrix
        aggregateAgesIndexes = np.arange(0, nBins + 1, int(nBins / len(snTypes)))
        confMatrixAggregateAges = get_aggregated_conf_matrix(aggregateAgesIndexes, testLabels, predictedLabels)
        classnames = np.copy(snTypes)
        if confMatrixAggregateAges.shape[0] < len(classnames):
            classnames = classnames[:-1]
        plot_confusion_matrix(confMatrixAggregateAges, classes=classnames, normalize=True, title='', fig_dir=fig_dir,
                              name='aggregate_ages', fontsize_labels=15, fontsize_matrix=16)

        # Aggregate age and subtypes conf matrix
        aggregateSubtypesIndexes = np.array([0, 108, 180, 234, 306])
        broadTypes = ['Ia', 'Ib', 'Ic', 'II']
        confMatrixAggregateSubtypes = get_aggregated_conf_matrix(aggregateSubtypesIndexes, testLabels, predictedLabels)
        plot_confusion_matrix(confMatrixAggregateSubtypes, classes=broadTypes, normalize=True, title='',
                              fig_dir=fig_dir, name='aggregate_subtypes', fontsize_labels=30, fontsize_matrix=30)
        # plt.show()

    np.set_printoptions(precision=2)
    print(confMatrix)
    plot_confusion_matrix(confMatrix, classes=typeNamesList, normalize=True, title='', fig_dir=fig_dir, name='all',
                          fontsize_labels=2, fontsize_matrix=1)

    # ACTUAL ACCURACY, broadTYPE ACCURACY, AGE ACCURACY
    typeAndAgeCorrect = 0
    typeCorrect = 0
    broadTypeCorrect = 0
    broadTypeAndAgeCorrect = 0
    typeAndNearAgeCorrect = 0
    broadTypeAndNearAgeCorrect = 0
    for i in range(len(testTypeNames)):
        predictedIndex = np.argmax(yy[i])

        classification = testTypeNames[i].split(': ')
        if len(classification) == 2:
            testType, testAge = classification
        else:
            testGalType, testType, testAge = classification
        actual = typeNamesList[predictedIndex].split(': ')
        if len(actual) == 2:
            actualType, actualAge = actual
        else:
            actualGalType, actualType, actualAge = actual

        testBroadType = testType[0:2]
        actualBroadType = actualType[0:2]
        if testType[0:3] == 'IIb':
            testBroadType = 'Ib'
        if actualType[0:3] == 'IIb':
            actualBroadType = 'Ib'
        nearTestAge = testAge.split(' to ')

        if testTypeNames[i] == typeNamesList[predictedIndex]:
            typeAndAgeCorrect += 1
        if testType == actualType:  # correct type
            typeCorrect += 1
            if (nearTestAge[0] in actualAge) or (
                    nearTestAge[1] in actualAge):  # check if the age is in the neigbouring bin
                typeAndNearAgeCorrect += 1  # all correct except nearby bin
        if testBroadType == actualBroadType:  # correct broadtype
            broadTypeCorrect += 1
            if testAge == actualAge:
                broadTypeAndAgeCorrect += 1
            if (nearTestAge[0] in actualAge) or (
                    nearTestAge[1] in actualAge):  # check if the age is in the neigbouring bin
                broadTypeAndNearAgeCorrect += 1  # Broadtype and nearby bin

    typeAndAgeAccuracy = float(typeAndAgeCorrect) / len(testTypeNames)
    typeAccuracy = float(typeCorrect) / len(testTypeNames)
    broadTypeAccuracy = float(broadTypeCorrect) / len(testTypeNames)
    broadTypeAndAgeAccuracy = float(broadTypeAndAgeCorrect) / len(testTypeNames)
    typeAndNearAgeAccuracy = float(typeAndNearAgeCorrect) / len(testTypeNames)
    broadTypeAndNearAgeAccuracy = float(broadTypeAndNearAgeCorrect) / len(testTypeNames)

    print("typeAndAgeAccuracy : " + str(typeAndAgeAccuracy))
    print("typeAccuracy : " + str(typeAccuracy))
    print("broadTypeAccuracy : " + str(broadTypeAccuracy))
    print("broadTypeAndAgeAccuracy: " + str(broadTypeAndAgeAccuracy))
    print("typeAndNearAgeAccuracy : " + str(typeAndNearAgeAccuracy))
    print("broadTypeAndNearAgeAccuracy : " + str(broadTypeAndNearAgeAccuracy))


def main():
    dirModel = "/Users/danmuth/PycharmProjects/astrodash/data_files_train80_splitspectra_zeroZ/"
    modelFilename = dirModel + "tensorflow_model.ckpt"

    fig_dir = os.path.join('..', 'Figures', 'zeroZ_train80_splitspectra')
    if not os.path.exists(fig_dir):
        os.makedirs(fig_dir)

    with open(os.path.join(dirModel, "training_params.pickle"), 'rb') as f:
        pars = pickle.load(f)
    snTypes = pars['typeList']

    dirTestSet = "/Users/danmuth/PycharmProjects/astrodash/data_files_train80_splitspectra_zeroZ/training_set/"
    testImagesAll = np.load(dirTestSet + 'testImages.npy', mmap_mode='r')
    testLabelsAll = np.load(dirTestSet + 'testLabels.npy', mmap_mode='r')
    typeNamesList = np.load(dirTestSet + 'typeNamesList.npy')
    testTypeNamesAll = np.load(dirTestSet + 'testTypeNames.npy')

    calc_model_metrics(modelFilename, testLabelsAll[:50000], testImagesAll[:50000], testTypeNamesAll[:50000],
                       typeNamesList, snTypes, fig_dir=fig_dir)


if __name__ == '__main__':
    main()



import numpy as np
from astropy.io import fits


def make_fits_file(inputFilename, outFilename):
    """ Code adapted from Mat Smith's 'spec_convert_global.py"""

    spec = np.loadtxt(inputFilename, dtype=[('wave', np.double), ('flux', np.double), ('err', np.double)])
    a = [np.where((spec['wave'] > 0) & (spec['wave'] < 200000))]
    prihdr = fits.Header()
    pix1 = a[0][0][0]
    pix2 = a[0][0][1]
    prihdr['CRVAL1'] = spec['wave'][pix1]
    prihdr['CDELT1'] = spec['wave'][pix2]-spec['wave'][pix1]
    prihdr['CRPIX1'] = 1.
    prihdr['CTYPE1'] = 'Wavelength'
    prihdr['CUNIT1'] = 'Angstroms'

    # The first image in the fits file is the flux
    hdu = fits.PrimaryHDU(spec['flux'][a], header=prihdr)
    hdulist = fits.HDUList(hdu)
    # The first extension is the variance: note the errors are squared: test this.
    hdulist.append(fits.ImageHDU((spec['err'][a])**2,name='variance'))
    # The next extension tells MarZ that we are looking at a transient.
    a1 = np.array(['DES16X3eww'])
    a2 = np.array(['P'])
    a3 = np.array(['Transient'])
    col1 = fits.Column(name='NAME', format='80A', array=a1)
    col2 = fits.Column(name='TYPE', format='1A', array=a2)
    col3 = fits.Column(name='COMMENT', format='80A', array=a3)
    cols = fits.ColDefs([col1,col2,col3])
    tbhdu = fits.BinTableHDU.from_columns(cols,name='FIBRES')
    hdulist.append(tbhdu)
    #  Finally we have the wavelength. This is in the header too, but it's good to heave it here for completeness.
    hdulist.append(fits.ImageHDU(spec['wave'][a],name='wavelength'))
    # Output it as a fits file
    hdulist.writeto(outFilename, overwrite=True)


if __name__ == "__main__":
    make_fits_file('/home/daniel/Documents/DES_Y4_Spectra/VLT/DES16C2aiy_all.txt', '/home/daniel/Desktop/outFits.fits')





import numpy as np
import pickle
import os
from astrodash.create_arrays import AgeBinning
from astrodash.helpers import temp_list
from astrodash.combine_sn_and_host import BinTemplate


def create_sn_and_host_arrays(snTemplateDirectory, snTempFileList, galTemplateDirectory, galTempFileList, paramsFile):
    snTemplates = {}
    galTemplates = {}
    snList = temp_list(snTempFileList)
    galList = temp_list(galTempFileList)
    with open(paramsFile, 'rb') as f:
        pars = pickle.load(f)
    w0, w1, nw, snTypes, galTypes, minAge, maxAge, ageBinSize = pars['w0'], pars['w1'], pars['nw'], pars['typeList'], \
                                                                pars['galTypeList'], pars['minAge'], pars['maxAge'], \
                                                                pars['ageBinSize']
    ageBinning = AgeBinning(minAge, maxAge, ageBinSize)
    ageLabels = ageBinning.age_labels()
    # Create dictionary of dictionaries for type and age of SN
    for snType in snTypes:
        snTemplates[snType] = {}
        for ageLabel in ageLabels:
            snTemplates[snType][ageLabel] = {}
            snTemplates[snType][ageLabel]['snInfo'] = []
            snTemplates[snType][ageLabel]['names'] = []
    for galType in galTypes:
        galTemplates[galType] = {}
        galTemplates[galType]['galInfo'] = []
        galTemplates[galType]['names'] = []

    for snFile in snList:
        snBinTemplate = BinTemplate(snTemplateDirectory + snFile, 'sn', w0, w1, nw)
        nAges = snBinTemplate.nCols
        ages = snBinTemplate.ages
        snType = snBinTemplate.tType
        filename = snBinTemplate.filename
        for ageIdx in range(nAges):
            age = ages[ageIdx]
            if minAge < age < maxAge:
                ageBin = ageBinning.age_bin(age)
                ageLabel = ageLabels[ageBin]
                snInfo = snBinTemplate.bin_template(ageIdx)
                snTemplates[snType][ageLabel]['snInfo'].append(snInfo)
                snTemplates[snType][ageLabel]['names'].append("%s_%s" % (filename, age))

            print("Reading {} {} out of {}".format(snFile, ageIdx, nAges))

    for galFile in galList:
        galBinTemplate = BinTemplate(galTemplateDirectory + galFile, 'gal', w0, w1, nw)
        galType = galBinTemplate.tType
        filename = galBinTemplate.filename
        galInfo = galBinTemplate.bin_template()
        galTemplates[galType]['galInfo'].append(galInfo)
        galTemplates[galType]['names'].append(filename)

        print("Reading {}".format(galFile))

    # Convert lists in dictionaries to numpy arrays
    for snType in snTypes:
        for ageLabel in ageLabels:
            snTemplates[snType][ageLabel]['snInfo'] = np.array(snTemplates[snType][ageLabel]['snInfo'])
            snTemplates[snType][ageLabel]['names'] = np.array(snTemplates[snType][ageLabel]['names'])
    for galType in galTypes:
        galTemplates[galType]['galInfo'] = np.array(galTemplates[galType]['galInfo'])
        galTemplates[galType]['names'] = np.array(galTemplates[galType]['names'])

    return snTemplates, galTemplates


def save_templates():
    scriptDirectory = os.path.dirname(os.path.abspath(__file__))
    parameterFile = 'models_v06/models/zeroZ/training_params.pickle'
    snTemplateDirectory = os.path.join(scriptDirectory, "../templates/training_set/")
    snTempFileList = snTemplateDirectory + 'templist.txt'
    galTemplateDirectory = os.path.join(scriptDirectory, "../templates/superfit_templates/gal/")
    galTempFileList = galTemplateDirectory + 'gal.list'
    saveFilename = 'models_v06/models/sn_and_host_templates.npz'

    snTemplates, galTemplates = create_sn_and_host_arrays(snTemplateDirectory, snTempFileList, galTemplateDirectory,
                                                          galTempFileList, parameterFile)

    np.savez_compressed(saveFilename, snTemplates=snTemplates, galTemplates=galTemplates)

    return saveFilename


if __name__ == "__main__":
    unCombinedTemplates = save_templates()



import os
import numpy as np
import astrodash

directoryPath = '/Users/dmuthukrishna/Documents/OzDES_data/ATEL_9742_Run26'

atel9742 = [
    ('DES16E1ciy_E1_combined_161101_v10_b00.dat', 0.174),
    ('DES16S1cps_S1_combined_161101_v10_b00.dat', 0.274),
    ('DES16E2crb_E2_combined_161102_v10_b00.dat', 0.229),
    ('DES16E2clk_E2_combined_161102_v10_b00.dat', 0.367),
    ('DES16E2cqq_E2_combined_161102_v10_b00.dat', 0.426),
    ('DES16X2ceg_X2_combined_161103_v10_b00.dat', 0.335),
    ('DES16X2bkr_X2_combined_161103_v10_b00.dat', 0.159),
    ('DES16X2crr_X2_combined_161103_v10_b00.dat', 0.312),
    ('DES16X2cpn_X2_combined_161103_v10_b00.dat', 0.28),
    ('DES16X2bvf_X2_combined_161103_v10_b00.dat', 0.135),
    ('DES16C1cbg_C1_combined_161103_v10_b00.dat', 0.111),
    ('DES16C2cbv_C2_combined_161103_v10_b00.dat', 0.109),
    ('DES16C1bnt_C1_combined_161103_v10_b00.dat', 0.351),
    ('DES16C3at_C3_combined_161031_v10_b00.dat', 0.217),
    ('DES16X3cpl_X3_combined_161031_v10_b00.dat', 0.205),
    ('DES16E2cjg_E2_combined_161102_v10_b00.dat', 0.48),
    ('DES16X2crt_X2_combined_161103_v10_b00.dat', 0.57)]

filenames = [os.path.join(directoryPath, i[0]) for i in atel9742]
knownRedshifts = [i[1] for i in atel9742]

classification = astrodash.Classify(filenames, knownRedshifts)
bestFits, bestTypes = classification.list_best_matches(n=1)
print(bestFits)
np.savetxt('Run26_fits.txt', bestFits, fmt='%s')
# classification.plot_with_gui(indexToPlot=1)




from scipy.fftpack import fft
from scipy.signal import argrelmax
import numpy as np
from scipy.stats import chisquare, pearsonr
from astrodash.restore_model import get_training_parameters
from astrodash.array_tools import mean_zero_spectra


# import matplotlib
# matplotlib.use('TkAgg')
# import matplotlib.pyplot as plt


def combined_prob(bestMatchList):
    hostName, prevName, age, prob = bestMatchList[0]
    probInitial = float(prob)
    bestName = prevName
    prevBroadTypeName = prevName[0:2]
    prevMinAge, prevMaxAge = age.split(' to ')
    probTotal = 0.
    agesList = [int(prevMinAge), int(prevMaxAge)]
    probPossible = 0.
    agesListPossible = []
    probPossible2 = 0.
    agesListPossible2 = []
    index = 0
    for host, name, age, prob in bestMatchList[0:10]:
        index += 1
        minAge, maxAge = list(map(int, age.split(' to ')))
        if "IIb" in name:
            broadTypeName = "Ib"
        else:
            broadTypeName = name[0:2]
        if name == prevName:
            if probPossible == 0:
                if (minAge in agesList) or (maxAge in agesList):
                    probTotal += float(prob)
                    prevName = name
                    agesList = agesList + [minAge, maxAge]
                else:
                    probPossible = float(prob)
                    agesListPossible = [minAge, maxAge]
            else:
                if probPossible2 == 0:
                    if ((minAge in agesListPossible) or (maxAge in agesListPossible)) and (
                            (minAge in agesList) or (maxAge in agesList)):
                        probTotal += probPossible + float(prob)
                        agesList = agesList + agesListPossible + [minAge, maxAge]
                        probPossible = 0
                        agesListPossible = []
                    else:
                        probPossible2 = probPossible + float(prob)
                        agesListPossible2 = agesListPossible + [minAge, maxAge]
                else:
                    if ((minAge in agesListPossible2) or (maxAge in agesListPossible2)) and (
                            (minAge in (agesList + agesListPossible)) or (maxAge in (agesList + agesListPossible))):
                        probTotal += probPossible2 + float(prob)
                        agesList = agesList + agesListPossible2 + [minAge, maxAge]
                        probPossible, probPossible2 = 0, 0
                        agesListPossible, agesListPossible2 = [], []

        elif broadTypeName == prevBroadTypeName:
            if probPossible == 0:
                if (minAge in agesList) or (maxAge in agesList):
                    if index <= 2:
                        bestName = broadTypeName
                    probTotal += float(prob)
                    agesList = agesList + [minAge, maxAge]
                else:
                    probPossible = float(prob)
                    agesListPossible = [minAge, maxAge]
            else:
                if ((minAge in agesListPossible) or (maxAge in agesListPossible)) and (
                        (minAge in agesList) or (maxAge in agesList)):
                    probTotal += probPossible + float(prob)
                    agesList = agesList + agesListPossible + [minAge, maxAge]
                    probPossible = 0
                    agesListPossible = []
                else:
                    break
        else:
            break

    bestAge = '%d to %d' % (min(agesList), max(agesList))

    if probTotal > probInitial:
        reliableFlag = True
    else:
        reliableFlag = False

    return hostName, bestName, bestAge, round(probTotal, 4), reliableFlag


class RlapCalc(object):
    def __init__(self, inputFlux, templateFluxes, templateNames, wave, inputMinMaxIndex, templateMinMaxIndexes):
        self.templateFluxes = templateFluxes
        self.templateNames = templateNames
        self.wave = wave
        pars = get_training_parameters()
        w0, w1, self.nw, = pars['w0'], pars['w1'], pars['nw']
        self.inputFlux = mean_zero_spectra(inputFlux, inputMinMaxIndex[0], inputMinMaxIndex[1], self.nw)
        self.templateMinMaxIndexes = templateMinMaxIndexes

        self.dwlog = np.log(w1 / w0) / self.nw

    def _cross_correlation(self, templateFlux, templateMinMaxIndex):
        templateFlux = mean_zero_spectra(templateFlux, templateMinMaxIndex[0], templateMinMaxIndex[1], self.nw)
        inputfourier = fft(self.inputFlux)
        tempfourier = fft(templateFlux)

        product = inputfourier * np.conj(tempfourier)
        xCorr = fft(product)

        rmsInput = np.std(inputfourier)
        rmsTemp = np.std(tempfourier)

        xCorrNorm = (1. / (self.nw * rmsInput * rmsTemp)) * xCorr

        rmsXCorr = np.std(product)

        xCorrNormRearranged = np.concatenate(
            (xCorrNorm[int(len(xCorrNorm) / 2):], xCorrNorm[0:int(len(xCorrNorm) / 2)]))

        crossCorr = np.correlate(self.inputFlux, templateFlux, mode='Full')[::-1][
                    int(self.nw / 2):int(self.nw + self.nw / 2)] / max(
            np.correlate(self.inputFlux, templateFlux, mode='Full'))

        try:
            deltapeak, h = self._get_peaks(crossCorr)[0]
            shift = int(deltapeak - self.nw / 2)
            autoCorr = np.correlate(templateFlux, templateFlux, mode='Full')[::-1][
                       int(self.nw / 2) - shift:int(self.nw + self.nw / 2) - shift] / max(
                np.correlate(templateFlux, templateFlux, mode='Full'))

            aRandomFunction = crossCorr - autoCorr
            rmsA = np.std(aRandomFunction)
        except IndexError as err:
            print("Error: Cross-correlation is zero, probably caused by empty spectrum.", err)
            rmsA = 1

        # import matplotlib.pyplot as plt
        # w0 = 3500.  # wavelength range in Angstroms
        # w1 = 10000.
        # nw = 1024.  # number of wavelength bins
        # dwlog = np.log(w1 / w0) / self.nw
        # zaxisindex1 = np.concatenate((np.arange(-self.nw / 2, 0), np.arange(0, self.nw / 2)))
        # zaxis1 = np.zeros(self.nw)
        # zaxis1[0:int(self.nw/2 -1)] = (np.exp(abs(zaxisindex1[0:int(self.nw/2-1)]) * dwlog) - 1)
        # zaxis1[int(self.nw/2):] = -(np.exp(abs(zaxisindex1[int(self.nw/2):]) * dwlog) - 1)
        # plt.figure(figsize=(10, 8))
        # plt.plot(zaxis1-0.24, xCorrNormRearranged, label='Cross-correlation, $c(n)$')
        # plt.plot(zaxis1-0.24, aRandomFunction/7, label='Antisymmetric component, $a(n)$', linestyle='--')
        # plt.axhline(rmsA/7, label='Antisymmetric rms, $\sigma_a$', color='k')
        # # plt.show()
        # # plt.plot(self.zAxis-0.24, np.correlate(self.inputFlux, templateFlux, mode='Full')[::-1][512:1536]/max(np.correlate(self.inputFlux, templateFlux, mode='Full')))
        # # plt.plot(self.zAxis-0.24, np.correlate(templateFlux, templateFlux, mode='Full')[::-1][512-shift:1536-shift]/max(np.correlate(templateFlux, templateFlux, mode='Full')))
        # plt.xlabel('$z - z_{peak}$', fontsize=20)
        # plt.ylabel('Normalised correlation amplitude', fontsize=20)
        # plt.legend(frameon=False, fontsize=15)
        # plt.xlim(-0.4, 0.4)
        # plt.tick_params(labelsize=18)
        # plt.tight_layout()
        # plt.savefig('/Users/danmuth/OneDrive/Documents/DASH/Paper/Figures/cross_correlation_{}.pdf'.format(np.random.randint(100)))
        # plt.show()

        return xCorr, rmsInput, rmsTemp, xCorrNorm, rmsXCorr, xCorrNormRearranged, rmsA

    def _get_peaks(self, crosscorr):
        peakindexes = argrelmax(crosscorr)[0]

        ypeaks = []
        for i in peakindexes:
            ypeaks.append(abs(crosscorr[i]))

        arr = list(zip(*[peakindexes, ypeaks]))
        arr.sort(key=lambda x: x[1])
        sortedPeaks = list(reversed(arr))

        return sortedPeaks

    def _calculate_r(self, crosscorr, rmsA):
        deltapeak1, h1 = self._get_peaks(crosscorr)[0]  # deltapeak = np.argmax(abs(crosscorr))
        deltapeak2, h2 = self._get_peaks(crosscorr)[1]
        # h = crosscorr[deltapeak]

        ##    shift = deltapeak
        ##    arms = 0
        ##    srms = 0
        ##    for k in range(0,int(nw-1)):
        ##        angle = -2*np.pi*k/nw
        ##        phase = complex(np.cos(angle),np.sin(angle))
        ####	f=2
        ####	if (k < k2):
        ####		arg = np.pi * float(k-k1)/float(k2-k1)
        ####		f = f * .25 * (1-np.cos(arg)) * (1-np.cos(arg))
        ####	elif (k > k3):
        ####		arg = np.pi * float(k-k3)/float(k4-k3)
        ####		f = f * .25 * (1+np.cos(arg)) * (1+np.cos(arg))
        ##        arms = arms + (phase*crosscorr[k+1]).imag*(phase*crosscorr[k+1]).imag
        ##        srms = srms + (phase*crosscorr[k+1]).real*(phase*crosscorr[k+1]).real
        ##
        ##    arms = np.sqrt(arms)/nw
        r = abs((h1 - rmsA) / (np.sqrt(2) * rmsA))
        fom = (h1 - 0.05) ** 0.75 * (h1 / h2)

        return r, deltapeak1, fom

    def get_redshift_axis(self, nw, dwlog):
        zAxisIndex = np.concatenate((np.arange(-nw / 2, 0), np.arange(0, nw / 2)))
        zAxis = np.zeros(nw)
        zAxis[0:int(nw / 2 - 1)] = -(np.exp(abs(zAxisIndex[0:int(nw / 2 - 1)]) * dwlog) - 1)
        zAxis[int(nw / 2):] = (np.exp(abs(zAxisIndex[int(nw / 2):]) * dwlog) - 1)
        zAxis = zAxis[::-1]

        return zAxis

    def calculate_rlap(self, crosscorr, rmsAntisymmetric, templateFlux):
        r, deltapeak, fom = self._calculate_r(crosscorr, rmsAntisymmetric)
        shift = int(deltapeak - self.nw / 2)  # shift from redshift

        # lap value
        iminindex, imaxindex = self.min_max_index(self.inputFlux)
        tminindex, tmaxindex = self.min_max_index(templateFlux)

        overlapminindex = int(max(iminindex + shift, tminindex))
        overlapmaxindex = int(min(imaxindex - 1 + shift, tmaxindex - 1))

        minWaveOverlap = self.wave[overlapminindex]
        maxWaveOverlap = self.wave[overlapmaxindex]

        lap = np.log(maxWaveOverlap / minWaveOverlap)
        rlap = 5 * r * lap

        fom = fom * lap
        # print r, lap, rlap, fom
        return r, lap, rlap, fom

    def min_max_index(self, flux):
        minindex, maxindex = (0, self.nw - 1)
        zeros = np.where(flux == 0)[0]
        j = 0
        for i in zeros:
            if (i != j):
                break
            j += 1
            minindex = j
        j = int(self.nw) - 1
        for i in zeros[::-1]:
            if (i != j):
                break
            j -= 1
            maxindex = j

        return minindex, maxindex

    def calculate_chi_squared(self, templateFlux):
        """ Only calculate on overlap region (MAYBE AVERAGE ALL TEMPLATES LATER)"""
        iminindex, imaxindex = self.min_max_index(self.inputFlux)
        tminindex, tmaxindex = self.min_max_index(templateFlux)

        overlapminindex = max(iminindex, tminindex)
        overlapmaxindex = min(imaxindex - 1, tmaxindex - 1)

        inputSpecOverlapped = 100 * (1 + self.inputFlux[overlapminindex:overlapmaxindex])
        templateSpecOverlapped = 100 * (1 + templateFlux[overlapminindex:overlapmaxindex])

        chi2 = chisquare(inputSpecOverlapped, templateSpecOverlapped)[0]
        pearsonCorr = pearsonr(inputSpecOverlapped, templateSpecOverlapped)

        return chi2, pearsonCorr

    def rlap_score(self, tempIndex):
        xcorr, rmsinput, rmstemp, xcorrnorm, rmsxcorr, xcorrnormRearranged, rmsA = self._cross_correlation(
            self.templateFluxes[tempIndex].astype('float'), self.templateMinMaxIndexes[tempIndex])
        crosscorr = xcorrnormRearranged
        r, lap, rlap, fom = self.calculate_rlap(crosscorr, rmsA, self.templateFluxes[tempIndex])

        return r, lap, rlap, fom

    def rlap_label(self):
        if not np.any(self.inputFlux):
            return "No flux", True

        self.zAxis = self.get_redshift_axis(self.nw, self.dwlog)
        rlapList = []

        for i in range(len(self.templateNames)):
            r, lap, rlap, fom = self.rlap_score(tempIndex=i)
            rlapList.append(rlap)
            if i > 20:
                break

        rlapMean = round(np.mean(rlapList), 2)
        rlapLabel = str(rlapMean)

        if rlapMean < 6:
            rlapWarning = True
        else:
            rlapWarning = False

        return rlapLabel, rlapWarning



import os
import glob
import shutil
import numpy as np
import itertools
import tensorflow as tf
from astrodash.multilayer_convnet import convnet_variables
import zipfile
import gzip
import time
from astrodash.array_tools import labels_indexes_to_arrays
from astrodash.model_metrics import calc_model_metrics
from astrodash.create_arrays import OverSampling
from astrodash.helpers import redshift_binned_spectrum, calc_params_for_log_redshifting


def train_model(dataDirName, overwrite=False, numTrainBatches=500000, minZ=0., maxZ=0., redshifting=False):
    """  Train model. Unzip and overwrite exisiting training set if overwrite is True"""
    # Open training data files
    trainingSet = os.path.join(dataDirName, 'training_set.zip')
    extractedFolder = os.path.join(dataDirName, 'training_set')
    # zipRef = zipfile.ZipFile(trainingSet, 'r')
    # zipRef.extractall(extractedFolder)
    # zipRef.close()
    if os.path.exists(extractedFolder) and overwrite:
        shutil.rmtree(extractedFolder, ignore_errors=True)
        os.system("unzip %s -d %s" % (trainingSet, extractedFolder))
    elif not os.path.exists(extractedFolder):
        os.system("unzip %s -d %s" % (trainingSet, extractedFolder))
    else:
        pass

    npyFiles = {}
    fileList = os.listdir(extractedFolder)
    for filename in fileList:
        f = os.path.join(extractedFolder, filename)
        if filename.endswith('.gz'):
            # # npyFiles[filename.strip('.npy.gz')] = gzip.GzipFile(f, 'r')
            # gzFile = gzip.open(f, "rb")
            # unCompressedFile = open(f.strip('.gz'), "wb")
            # decoded = gzFile.read()
            # unCompressedFile.write(decoded)
            # gzFile.close()
            # unCompressedFile.close()
            npyFiles[filename.strip('.npy.gz')] = f.strip('.gz')
            os.system("gzip -d %s" % f)  # "gzip -dk %s" % f
        elif filename.endswith('.npy'):
            npyFiles[filename.strip('.npy')] = f

    trainImages = np.load(npyFiles['trainImages'], mmap_mode='r')
    trainLabels = np.load(npyFiles['trainLabels'], mmap_mode='r')
    testImages = np.load(npyFiles['testImages'], mmap_mode='r')
    testLabelsIndexes = np.load(npyFiles['testLabels'], mmap_mode='r')
    typeNamesList = np.load(npyFiles['typeNamesList'])
    testTypeNames = np.load(npyFiles['testTypeNames'])
    # testImages = np.load(npyFiles['testImagesNoGal'], mmap_mode='r')
    # testLabelsIndexes = np.load(npyFiles['testLabelsNoGal'], mmap_mode='r')

    print("Completed creatingArrays")
    print(len(trainImages))

    nLabels = len(typeNamesList)
    N = 1024
    nIndexes, dwlog, w0, w1, nw = calc_params_for_log_redshifting(dataDirName)

    overSampling = OverSampling(nLabels, N, images=trainImages, labels=trainLabels)
    trainArrays = overSampling.over_sample_arrays(smote=False)
    trainImages, trainLabels = trainArrays['images'], trainArrays['labels']

    # # Delete temporary memory mapping files
    # for filename in glob.glob('shuffled*.dat') + glob.glob('oversampled*.dat'):
    #     if not os.path.samefile(filename, trainImages.filename) and not os.path.samefile(filename, trainLabels.filename):
    #         os.remove(filename)

    # Set up the convolutional network architecture
    imWidth = 32  # Image size and width
    imWidthReduc = 8
    a = []
    x, y_, keep_prob, y_conv, W, b = convnet_variables(imWidth, imWidthReduc, N, nLabels)

    with tf.Session() as sess:  # config=tf.ConfigProto(inter_op_parallelism_threads=1, intra_op_parallelism_threads=1)) as sess:
        # TRAIN AND EVALUATE MODEL
        cross_entropy = tf.reduce_mean(-tf.reduce_sum(y_ * tf.log(y_conv + 1e-8), reduction_indices=[1]))
        train_step = tf.train.AdamOptimizer(1e-4).minimize(cross_entropy)
        correct_prediction = tf.equal(tf.argmax(y_conv, 1), tf.argmax(y_, 1))
        accuracy = tf.reduce_mean(tf.cast(correct_prediction, tf.float32))

        sess.run(tf.global_variables_initializer())

        testLabels = labels_indexes_to_arrays(testLabelsIndexes[0:400], nLabels)
        testImages = testImages[0:400]
        # testLabelsWithGal = labels_indexes_to_arrays(testLabelsIndexesWithGal[0:200], nLabels)
        # testImagesWithGal = testImagesWithGal[0:200]

        trainImagesCycle = itertools.cycle(trainImages)
        trainLabelsCycle = itertools.cycle(trainLabels)
        for i in range(numTrainBatches):
            batch_xs = np.array(list(itertools.islice(trainImagesCycle, 50 * i, 50 * i + 50)))
            batch_ys = labels_indexes_to_arrays(list(itertools.islice(trainLabelsCycle, 50 * i, 50 * i + 50)), nLabels)

            # Redshift arrays
            if redshifting is True:
                redshifts = np.random.uniform(low=minZ, high=maxZ, size=len(batch_xs))
                for j, z in enumerate(redshifts):
                    batch_xs[j] = redshift_binned_spectrum(batch_xs[j], z, nIndexes, dwlog, w0, w1, nw, outerVal=0.5)

            train_step.run(feed_dict={x: batch_xs, y_: batch_ys, keep_prob: 0.5})
            if i % 100 == 0:
                train_accuracy = accuracy.eval(feed_dict={x: batch_xs, y_: batch_ys, keep_prob: 1.0})
                print("step %d, training accuracy %g" % (i, train_accuracy))
                testAcc = accuracy.eval(feed_dict={x: testImages, y_: testLabels, keep_prob: 1.0})
                print("test accuracy %g" % testAcc)
                a.append(testAcc)
                # if i % 1000 == 0:
                #     testWithGalAcc = accuracy.eval(feed_dict={x: testImagesWithGal, y_: testLabelsWithGal, keep_prob: 1.0})
                #     print("test With Gal accuracy %g" % testWithGalAcc)

        print("test accuracy %g" % accuracy.eval(feed_dict={x: testImages, y_: testLabels, keep_prob: 1.0}))

        yy = y_conv.eval(feed_dict={x: testImages, y_: testLabels, keep_prob: 1.0})
        cp = correct_prediction.eval(feed_dict={x: testImages, y_: testLabels, keep_prob: 1.0})
        print(cp)
        for i in range(len(cp)):
            if cp[i] == False:
                predictedIndex = np.argmax(yy[i])
                print(i, testTypeNames[i], typeNamesList[predictedIndex])

        # SAVE THE MODEL
        saveFilename = os.path.join(dataDirName, "tensorflow_model.ckpt")
        saver = tf.train.Saver()
        save_path = saver.save(sess, saveFilename)
        print("Model saved in file: %s" % save_path)

    modelFilenames = [saveFilename + '.index', saveFilename + '.meta', saveFilename + '.data-00000-of-00001']

    try:
        import matplotlib.pyplot as plt
        plt.plot(a)
        plt.xlabel("Number of Epochs")
        plt.ylabel("Testing accuracy")
        plt.savefig(os.path.join(dataDirName, "testing_accuracy.png"))
        np.savetxt(os.path.join(dataDirName, "testing_accuracy.txt"), np.array(a))
    except Exception as e:
        print(e)

    try:
        calc_model_metrics(saveFilename, testLabelsIndexes, testImages, testTypeNames, typeNamesList)
    except Exception as e:
        print(e)

    return modelFilenames


if __name__ == '__main__':
    t1 = time.time()
    savedFilenames = train_model('data_files_zeroZ/', overwrite=False, numTrainBatches=500000, redshifting=True, minZ=0,
                                 maxZ=0.8)
    t2 = time.time()
    print("time spent: {0:.2f}".format(t2 - t1))



import os
import astrodash


def read_ozdes_wiki_atel(atelTextFile):
    """ Read ATEL wiki text, which has been saved to a text file: atelTextFile. 
    Returns the names, redshifts, and wikiClassifications"""
    names = []
    redshifts = []
    wikiClassifications = []
    with open(atelTextFile, 'r') as f:
        for line in f:
            if line[0:3] == 'DES':
                objectInfo = line.split()
                name = objectInfo[0]
                redshift = objectInfo[1]
                wikiClassification = ' '.join(objectInfo[2:])

                try:
                    if redshift != '?':  #  and 'SN' not in wikiClassification:
                        names.append(name)
                        redshifts.append(float(redshift))
                        wikiClassifications.append(wikiClassification)
                except ValueError as e:
                    print("Invalid redshift for line: {0}".format(line))
                    raise e

    return names, redshifts, wikiClassifications


def main(runDirectory, atelTextFile, saveMatchesFilename):

    # Store all the file paths to the objects in this run
    directoryPath = os.path.join(os.path.dirname(os.path.abspath(__file__)), runDirectory)
    allFilePaths = []
    for dirpath, dirnames, filenames in os.walk(directoryPath):
        for filename in [f for f in filenames if f.endswith(".dat")]:
            allFilePaths.append(os.path.join(dirpath, filename))
    allFilePaths.reverse()

    # Get filenames and corresponding redshifts
    names, knownRedshifts, wikiClassifications = read_ozdes_wiki_atel(atelTextFile)
    run = []
    for i in range(len(names)):
        for filePath in allFilePaths:
            if names[i] == filePath.split('/')[-1].split('_')[0]:
                run.append((filePath, knownRedshifts[i], wikiClassifications[i]))
                #break  # Uncomment the break to only classify the last dated spectrum for each object instead of classifying all dates.
    filenames = [i[0] for i in run]
    knownRedshifts = [i[1] for i in run]
    wikiClassifications = [i[2] for i in run]

    # Classify and print the best matches
    classification = astrodash.Classify(filenames, knownRedshifts, classifyHost=False, rlapScores=True, smooth=6)
    bestFits, redshifts, bestTypes, rlapFlag, matchesFlag, redshiftErrs = classification.list_best_matches(n=5, saveFilename=saveMatchesFilename)
    print("{0:17} | {1:5} | {2:8} | {3:10} | {4:6} | {5:10} | {6:10}".format("Name", "  z  ", "DASH_Fit", "  Age ", "Prob.", "Flag", "Wiki Fit"))
    for i in range(len(filenames)):
        print("{0:17} | {1:5} | {2:8} | {3:10} | {4:6} | {5:10} | {6:10}".format('_'.join([filenames[i].split('/')[-1].split('_')[0], filenames[i].split('/')[-1].split('_')[3]]) , redshifts[i], bestTypes[i][0], bestTypes[i][1], bestTypes[i][2], matchesFlag[i].replace(' matches',''), wikiClassifications[i]))

    # Plot one of the matches
    classification.plot_with_gui(indexToPlot=0)


if __name__ == '__main__':
    main(runDirectory='../templates/run034/', atelTextFile='wiki_atel_run034.txt', saveMatchesFilename='DASH_matches_run34.txt')



import numpy as np
import os
from astrodash.combine_sn_and_host import CombineSnAndHost

scriptDirectory = os.path.dirname(os.path.abspath(__file__))


def get_templates(snName, snAge, hostName, snTemplates, galTemplates, nw):
    snInfos = np.copy(snTemplates[snName][snAge]['snInfo'])
    snNames = np.copy(snTemplates[snName][snAge]['names'])
    if hostName != "No Host" and hostName != "":
        hostInfos = np.copy(galTemplates[hostName]['galInfo'])
        hostNames = np.copy(galTemplates[hostName]['names'])
    else:
        hostInfos = np.array([[np.zeros(nw), np.zeros(nw), 1, nw - 1]])
        hostNames = np.array(["No Host"])

    return snInfos, snNames, hostInfos, hostNames


def load_templates(templateFilename):
    loaded = np.load(os.path.join(scriptDirectory, templateFilename), allow_pickle=True)
    snTemplates = loaded['snTemplates'][()]
    galTemplates = loaded['galTemplates'][()]

    return snTemplates, galTemplates


def combined_sn_and_host_data(snCoeff, galCoeff, z, snInfo, galInfo, w0, w1, nw):
    combineSnAndHost = CombineSnAndHost(snInfo, galInfo, w0, w1, nw)

    return combineSnAndHost.template_data(snCoeff, galCoeff, z)


if __name__ == "__main__":
    templateFilename1 = 'models/sn_and_host_templates.npz'
    snTemplates1, galTemplates1 = load_templates(templateFilename1)
    snInfoList = snTemplates1['Ia-norm']['-2 to 2']['snInfo']
    galInfoList = galTemplates1['S0']['galInfo']
    for i in range(len(snInfoList)):
        wave, flux, minMaxIndex = combined_sn_and_host_data(snCoeff=0.5, galCoeff=0.5, z=0, snInfo=snInfoList[i],
                                                            galInfo=galInfoList[0], w0=3500, w1=10000, nw=1024)
        print(i)



import astrodash
import os


directory = '/Users/danmuth/PycharmProjects/DASH/templates/redesgeminireductions/'


names = ["des17x3rg", "des17x1kw", "des17x3b", "des17x3t", "des17c3e", "des17c3w", "des17c1azd", "des17x2blx",
        "des17s2bzl", "des17s1bqr", "des17c3bzw", "des17x1cu"]

redshifts = [0.280, 0.462, 0.67, 0.49, 0.260, 0.675, 0.341, 0.344, 0.651, 0.649, 0.768, 0.83]

filenames = os.listdir(directory)


fnames = []
z = []
for i, name in enumerate(names):
    for filename in filenames:
        if name in filename:
            fnames.append(directory + filename)
            z.append(redshifts[i])

classification = astrodash.Classify(fnames, z, classifyHost=False, smooth=5, knownZ=True)
bestFits, redshifts, bestTypes, rlapFlag, matchesFlag, redshiftErrs = classification.list_best_matches(n=5)

print(bestFits)
f = open('gemini_classifications.txt', 'w')
for i in range(len(fnames)):
    f.write("%s   z=%s     %s      %s     %s\n %s\n\n" % (fnames[i].replace(directory, ''), z[i], bestTypes[i], matchesFlag[i], rlapFlag[i], bestFits[i]))
f.close()
print("Finished classifying %d spectra!" % len(fnames))


print("{0:17} | {1:5} | {2:8} | {3:10} | {4:6} | {5:10}".format("Name", "  z  ", "DASH_Fit", "  Age ", "Prob.", "Flag"))
for i in range(len(fnames)):
    print("{0:17} | {1:5} | {2:8} | {3:10} | {4:6} | {5:10}".format(
        fnames[i].replace(directory, ''), z[i],
        bestTypes[i][0], bestTypes[i][1], bestTypes[i][2], matchesFlag[i].replace(' matches', '')))

# PLOT SPECTRUM ON GUI
classification.plot_with_gui(indexToPlot=0)



import sys
import argparse
from astrodash.download_data_files import download_all_files

download_all_files('v06')

try:
    from PyQt5 import QtGui
    from astrodash.gui_main import MainApp
except ImportError:
    print("Warning: You will need to install 'PyQt5' if you want to use the graphical interface. " \
          "Using the automatic library will continue to work.")

from astrodash.classify import Classify


def main():
    classification = Classify(filenames=['test_spectrum_file.dat',
                                         'test_spectrum_file.dat'],
                              redshifts=[0.34, 0.13])
    print(classification.list_best_matches())
    # classification.plot_with_gui(indexToPlot=1)


def run_gui():
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', "--filepath", type=str, help='Spectrum filepath (str)')
    parser.add_argument('-z', "--redshift", type=float, help='Spectrum redshift (float)')
    parser.add_argument('-g', "--classify_host", help='Classify Host', action='store_true')
    parser.add_argument('-s', "--smooth", type=int, help='Smooth spectrum (int)')
    args = parser.parse_args()

    app = QtGui.QApplication(sys.argv)

    if args.filepath is None:
        filepath = "DefaultFilename"
    else:
        filepath = args.filepath
    form = MainApp(inputFilename=filepath)

    if args.redshift is None:
        knownZ = False
        redshift = ""
    else:
        knownZ = True
        redshift = str(args.redshift)
    form.checkBoxKnownZ.setChecked(knownZ)
    form.lineEditKnownZ.setText(str(redshift))

    if args.classify_host:
        form.checkBoxClassifyHost.setChecked(True)
        form.classifyHost = True

    if args.smooth is not None:
        form.lineEditSmooth.setText(str(args.smooth))

    if args.filepath is not None:
        form.lineEditInputFilename.setText(filepath)
        form.select_tensorflow_model()
        form.fit_spectra()
    form.show()
    app.exec_()


if __name__ == '__main__':
    main()
    # run_gui()





import astrodash
import numpy as np

wave = np.arange(2500, 10000, 2)
flux = np.ones(len(wave))
filename = np.array([wave, flux])
# filename = open('/Users/danmuth/PycharmProjects/DASH/templates/superfit_templates/sne/Ib/sn1983N.p12.dat', 'r')
# filename = 'osc-SN2002er-10'
redshift = 0

classification = astrodash.Classify([filename], [redshift], classifyHost=False, knownZ=True, smooth=6, rlapScores=True)
bestFits, redshifts, bestTypes, rlapFlag, matchesFlag, redshiftErrs = classification.list_best_matches(n=5)

print(bestFits)

classification.plot_with_gui(indexToPlot=0)
